# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

from __future__ import division

import warnings
import numpy as np
import math
import logging

from hyperspy.misc.utils import stash_active_state
from exspy.misc.eds.utils import _get_element_and_line

from hyperspy.models.model1d import Model1D
from exspy.signals.eds import EDSSpectrum
from exspy.misc.elements import elements as elements_db
from exspy.misc.eds import utils as utils_eds
import hyperspy.components1d as create_component

_logger = logging.getLogger(__name__)

eV2keV = 1000.0
sigma2fwhm = 2 * math.sqrt(2 * math.log(2))


def _get_weight(element, line, weight_line=None):
    if weight_line is None:
        weight_line = elements_db[element]["Atomic_properties"]["Xray_lines"][line][
            "weight"
        ]
    return "x * {}".format(weight_line)


def _get_sigma(E, E_ref, units_factor, return_f=False):
    """
    Calculates an approximate sigma value, accounting for peak broadening due
    to the detector, for a peak at energy E given a known width at a reference
    energy.

    The factor 2.5 is a constant derived by Fiori & Newbury as references
    below.

    Parameters
    ----------
    energy_resolution_MnKa : float
        Energy resolution of Mn Ka in eV
    E : float
        Energy of the peak in keV

    Returns
    -------
    float : FWHM of the peak in keV

    Notes
    -----
    This method implements the equation derived by Fiori and Newbury as is
    documented in the following:

        Fiori, C. E., and Newbury, D. E. (1978). In SEM/1978/I, SEM, Inc.,
        AFM O'Hare, Illinois, p. 401.

        Goldstein et al. (2003). "Scanning Electron Microscopy & X-ray
        Microanalysis", Plenum, third edition, p 315.
    """
    energy2sigma_factor = 2.5 / (eV2keV * (sigma2fwhm**2))
    if return_f:
        return lambda sig_ref: math.sqrt(
            abs(energy2sigma_factor * (E - E_ref) * units_factor + np.power(sig_ref, 2))
        )
    else:
        return "sqrt(abs({} * ({} - {}) * {} + sig_ref ** 2))".format(
            energy2sigma_factor, E, E_ref, units_factor
        )


def _get_offset(diff):
    return "x + {}".format(diff)


def _get_scale(E1, E_ref1, fact):
    return "{} + {} * (x - {})".format(E1, fact, E_ref1)


class EDSModel(Model1D):

    """Build and fit a model of an EDS Signal1D.

    Parameters
    ----------
    spectrum : EDSSpectrum (or any EDSSpectrum subclass) instance.

    auto_add_lines : bool
        If True, automatically add Gaussians for all X-rays generated
        in the energy range by an element, using the edsmodel.add_family_lines
        method.
    auto_background : bool
        If True, adds automatically a polynomial order 6 to the model,
        using the edsmodel.add_polynomial_background method.

    Any extra arguments are passed to the Model creator.

    Example
    -------
    >>> m = s.create_model()
    >>> m.fit()
    >>> m.fit_background()
    >>> m.calibrate_energy_axis('resolution')
    >>> m.calibrate_xray_lines('energy', ['Au_Ma'])
    >>> m.calibrate_xray_lines('sub_weight',['Mn_La'], bound=10)
    """

    def __init__(
        self, spectrum, auto_background=True, auto_add_lines=True, *args, **kwargs
    ):
        Model1D.__init__(self, spectrum, *args, **kwargs)
        self.xray_lines = list()
        self.family_lines = list()
        end_energy = self.axes_manager.signal_axes[0].high_value
        self.end_energy = min(end_energy, self.signal._get_beam_energy())
        self.start_energy = self.axes_manager.signal_axes[0].low_value
        self.background_components = list()
        if "dictionary" in kwargs or len(args) > 1:
            auto_add_lines = False
            auto_background = False
            d = args[1] if len(args) > 1 else kwargs["dictionary"]
            if len(d["xray_lines"]) > 0:
                self.xray_lines.extend([self[name] for name in d["xray_lines"]])
            if len(d["background_components"]) > 0:
                self.background_components.extend(
                    [self[name] for name in d["background_components"]]
                )
        if auto_background is True:
            self.add_polynomial_background()
        if auto_add_lines is True:
            # Will raise an error if no elements are specified, so check:
            if "Sample.elements" in self.signal.metadata:
                self.add_family_lines()

    def as_dictionary(self, fullcopy=True):
        dic = super(EDSModel, self).as_dictionary(fullcopy)
        dic["xray_lines"] = [c.name for c in self.xray_lines]
        dic["background_components"] = [c.name for c in self.background_components]
        return dic

    @property
    def units_factor(self):
        units_name = self.axes_manager.signal_axes[0].units
        if units_name == "eV":
            return 1000.0
        elif units_name == "keV":
            return 1.0
        else:
            raise ValueError("Energy units, %s, not supported" % str(units_name))

    @property
    def spectrum(self):
        return self._signal

    @spectrum.setter
    def spectrum(self, value):
        if isinstance(value, EDSSpectrum):
            self._signal = value
        else:
            raise ValueError(
                "This attribute can only contain an EDSSpectrum "
                "but an object of type %s was provided" % str(type(value))
            )

    @property
    def _active_xray_lines(self):
        return [xray_line for xray_line in self.xray_lines if xray_line.active]

    def add_family_lines(self, xray_lines="from_elements"):
        """Create the Xray-lines instances and configure them appropiately

        If a X-ray line is given, all the the lines of the familiy is added.
        For instance if Zn Ka is given, Zn Kb is added too. The main lines
        (alpha) is added to self.xray_lines

        Parameters
        -----------
        xray_lines: {None, 'from_elements', list of string}
            If None, if `metadata` contains `xray_lines` list of lines use
            those. If 'from_elements', add all lines from the elements contains
            in `metadata`. Alternatively, provide an iterable containing
            a list of valid X-ray lines symbols. (eg. ('Al_Ka','Zn_Ka')).

        Raises
        ------
        NotImplementedError
            If the signal axis is a non-uniform axis.
        """
        # Test that signal axis is uniform
        if not self.axes_manager[-1].is_uniform:
            raise NotImplementedError(
                "This function is not yet implemented " "for non-uniform axes."
            )

        only_one = False
        only_lines = ("Ka", "La", "Ma")

        if xray_lines is None or xray_lines == "from_elements":
            if (
                "Sample.xray_lines" in self.signal.metadata
                and xray_lines != "from_elements"
            ):
                xray_lines = self.signal.metadata.Sample.xray_lines
            elif "Sample.elements" in self.signal.metadata:
                xray_lines = self.signal._get_lines_from_elements(
                    self.signal.metadata.Sample.elements,
                    only_one=only_one,
                    only_lines=only_lines,
                )
            else:
                raise ValueError("No elements defined, set them with `add_elements`")

        components_names = [xr.name for xr in self.xray_lines]
        xray_lines = filter(lambda x: x not in components_names, xray_lines)
        xray_lines, xray_not_here = self.signal._get_xray_lines_in_spectral_range(
            xray_lines
        )
        for xray in xray_not_here:
            warnings.warn("%s is not in the data energy range." % (xray))

        for xray_line in xray_lines:
            element, line = utils_eds._get_element_and_line(xray_line)
            line_energy, line_FWHM = self.signal._get_line_energy(
                xray_line, FWHM_MnKa="auto"
            )
            component = create_component.Gaussian()
            component.centre.value = line_energy
            component.fwhm = line_FWHM
            component.centre.free = False
            component.sigma.free = False
            component.name = xray_line
            self.append(component)
            self.xray_lines.append(component)
            if self.signal._lazy:
                # For lazy signal, use a default value to avoid having
                # to do out-of-core computation
                component.A.map["values"] = 10
            else:
                component.A.map["values"] = (
                    self.signal.isig[line_energy].data
                    * line_FWHM
                    / self.signal.axes_manager[-1].scale
                )
            component.A.map["is_set"] = True
            component.A.ext_force_positive = True
            for li in elements_db[element]["Atomic_properties"]["Xray_lines"]:
                if line[0] in li and line != li:
                    xray_sub = element + "_" + li
                    if (
                        self.signal._get_xray_lines_in_spectral_range([xray_sub])[0]
                        != []
                    ):
                        line_energy, line_FWHM = self.signal._get_line_energy(
                            xray_sub, FWHM_MnKa="auto"
                        )
                        component_sub = create_component.Gaussian()
                        component_sub.centre.value = line_energy
                        component_sub.fwhm = line_FWHM
                        component_sub.centre.free = False
                        component_sub.sigma.free = False
                        component_sub.name = xray_sub
                        component_sub.A.twin_function_expr = _get_weight(element, li)
                        component_sub.A.twin = component.A
                        self.append(component_sub)
                        self.family_lines.append(component_sub)
            self.fetch_stored_values()

    @property
    def _active_background_components(self):
        return [bc for bc in self.background_components if bc.free_parameters]

    def add_polynomial_background(self, order=6):
        """
        Add a polynomial background.

        the background is added to self.background_components

        Parameters
        ----------
        order: int
            The order of the polynomial
        """
        background = create_component.Polynomial(order=order)
        background.name = "background_order_" + str(order)
        background.isbackground = True
        self.append(background)
        self.background_components.append(background)

    def free_background(self):
        """
        Free the yscale of the background components.
        """
        for component in self.background_components:
            component.set_parameters_free()

    def fix_background(self):
        """
        Fix the background components.
        """
        for component in self._active_background_components:
            component.set_parameters_not_free()

    def enable_xray_lines(self):
        """Enable the X-ray lines components."""
        for component in self.xray_lines:
            component.active = True

    def disable_xray_lines(self):
        """Disable the X-ray lines components."""
        for component in self._active_xray_lines:
            component.active = False

    def _make_position_adjuster(self, component, fix_it, show_label):
        # Override to ensure formatting of labels of xray lines
        super(EDSModel, self)._make_position_adjuster(component, fix_it, show_label)
        if show_label and component in (self.xray_lines + self.family_lines):
            label = self._position_widgets[component._position][1]
            label.string = r"$\mathrm{%s}_{\mathrm{%s}}$" % _get_element_and_line(
                component.name
            )

    def fit_background(
        self,
        start_energy=None,
        end_energy=None,
        windows_sigma=(4.0, 3.0),
        kind="single",
        **kwargs,
    ):
        """
        Fit the background in the energy range containing no X-ray line.

        After the fit, the background is fixed.

        Parameters
        ----------
        start_energy : {float, None}
            If float, limit the range of energies from the left to the
            given value.
        end_energy : {float, None}
            If float, limit the range of energies from the right to the
            given value.
        windows_sigma: tuple of two float
            The (lower, upper) bounds around each X-ray line, each as a float,
            to define the energy range free of X-ray lines.
        kind : {'single', 'multi'}
            If 'single' fit only the current location. If 'multi'
            use multifit.
        **kwargs : extra key word arguments
            All extra key word arguments are passed to fit or multifit

        See also
        --------
        free_background
        """

        if end_energy is None:
            end_energy = self.end_energy
        if start_energy is None:
            start_energy = self.start_energy

        # disactivate line
        self.free_background()
        with stash_active_state(self):
            self.disable_xray_lines()
            self.set_signal_range(start_energy, end_energy)
            for component in self:
                if component.isbackground is False:
                    self.remove_signal_range(
                        component.centre.value
                        - windows_sigma[0] * component.sigma.value,
                        component.centre.value
                        + windows_sigma[1] * component.sigma.value,
                    )
            if kind == "single":
                self.fit(**kwargs)
            if kind == "multi":
                self.multifit(**kwargs)
            self.reset_signal_range()
        self.fix_background()

    def _twin_xray_lines_width(self, xray_lines):
        """
        Twin the width of the peaks

        The twinning models the energy resolution of the detector

        Parameters
        ----------
        xray_lines: list of str or 'all_alpha'
            The Xray lines. If 'all_alpha', fit all using all alpha lines
        """
        if xray_lines == "all_alpha":
            xray_lines = [compo.name for compo in self.xray_lines]

        for i, xray_line in enumerate(xray_lines):
            component = self[xray_line]
            if i == 0:
                component_ref = component
                component_ref.sigma.free = True
                E_ref = component_ref.centre.value
            else:
                component.sigma.free = True
                E = component.centre.value
                component.sigma.twin_inverse_function_expr = _get_sigma(
                    E_ref, E, self.units_factor
                )
                component.sigma.twin_function_expr = _get_sigma(
                    E, E_ref, self.units_factor
                )

    def _set_energy_resolution(self, xray_lines, *args, **kwargs):
        """
        Adjust the width of all lines and set the fitted energy resolution
        to the spectrum

        Parameters
        ----------
        xray_lines: list of str or 'all_alpha'
            The Xray lines. If 'all_alpha', fit all using all alpha lines
        """
        if xray_lines == "all_alpha":
            xray_lines = [compo.name for compo in self.xray_lines]
        energy_Mn_Ka, FWHM_MnKa_old = self.signal._get_line_energy("Mn_Ka", "auto")
        FWHM_MnKa_old *= eV2keV / self.units_factor
        get_sigma_Mn_Ka = _get_sigma(
            energy_Mn_Ka,
            self[xray_lines[0]].centre.value,
            self.units_factor,
            return_f=True,
        )
        FWHM_MnKa = (
            get_sigma_Mn_Ka(self[xray_lines[0]].sigma.value)
            * eV2keV
            / self.units_factor
            * sigma2fwhm
        )
        if FWHM_MnKa < 110:
            raise ValueError(
                "FWHM_MnKa of "
                + str(FWHM_MnKa)
                + " smaller than"
                + "physically possible"
            )
        else:
            self.signal.set_microscope_parameters(energy_resolution_MnKa=FWHM_MnKa)
            _logger.info(
                "Energy resolution (FWHM at Mn Ka) changed from "
                + "{:.2f} to {:.2f} eV".format(FWHM_MnKa_old, FWHM_MnKa)
            )
            for component in self:
                if component.isbackground is False:
                    line_FWHM = self.signal._get_line_energy(
                        component.name, FWHM_MnKa="auto"
                    )[1]
                    component.fwhm = line_FWHM

    def _twin_xray_lines_scale(self, xray_lines):
        """
        Twin the scale of the peaks

        Parameters
        ----------
        xray_lines: list of str or 'all_alpha'
            The Xray lines. If 'all_alpha', fit all using all alpha lines
        """
        if xray_lines == "all_alpha":
            xray_lines = [compo.name for compo in self.xray_lines]
        ax = self.signal.axes_manager[-1]
        ref = []
        for i, xray_line in enumerate(xray_lines):
            component = self[xray_line]
            if i == 0:
                component_ref = component
                component_ref.centre.free = True
                E_ref = component_ref.centre.value
                ref.append(E_ref)
            else:
                component.centre.free = True
                E = component.centre.value
                fact = float(ax.value2index(E)) / ax.value2index(E_ref)
                component.centre.twin_function_expr = _get_scale(E, E_ref, fact)
                component.centre.twin = component_ref.centre
                ref.append(E)
        return ref

    def _set_energy_scale(self, xray_lines, ref):
        """
        Adjust the width of all lines and set the fitted energy resolution
        to the spectrum

        Parameters
        ----------
        xray_lines: list of str or 'all_alpha'
            The X-ray lines. If 'all_alpha', fit all using all alpha lines
        ref: list of float
            The centres, before fitting, of the X-ray lines included

        Raises
        ------
        NotImplementedError
            If the signal axis is a non-uniform axis.
        """
        # Test that signal axis is uniform
        if not self.axes_manager[-1].is_uniform:
            raise NotImplementedError(
                "This function is not yet implemented " "for non-uniform axes."
            )

        if xray_lines == "all_alpha":
            xray_lines = [compo.name for compo in self.xray_lines]
        ax = self.signal.axes_manager[-1]
        scale_old = self.signal.axes_manager[-1].scale
        ind = np.argsort(np.array([compo.centre.value for compo in self.xray_lines]))[
            -1
        ]
        E = self[xray_lines[ind]].centre.value
        scale = (ref[ind] - ax.offset) / ax.value2index(E)
        ax.scale = scale
        for i, xray_line in enumerate(xray_lines):
            component = self[xray_line]
            component.centre.value = ref[i]
        _logger.info("Scale changed from  %lf to %lf", scale_old, scale)

    def _twin_xray_lines_offset(self, xray_lines):
        """
        Twin the offset of the peaks

        Parameters
        ----------
        xray_lines: list of str or 'all_alpha'
            The Xray lines. If 'all_alpha', fit all using all alpha lines
        """
        if xray_lines == "all_alpha":
            xray_lines = [compo.name for compo in self.xray_lines]
        ref = []
        for i, xray_line in enumerate(xray_lines):
            component = self[xray_line]
            if i == 0:
                component_ref = component
                component_ref.centre.free = True
                E_ref = component_ref.centre.value
                ref.append(E_ref)
            else:
                component.centre.free = True
                E = component.centre.value
                diff = E_ref - E
                component.centre.twin_function_expr = _get_offset(-diff)
                component.centre.twin = component_ref.centre
                ref.append(E)
        return ref

    def _set_energy_offset(self, xray_lines, ref):
        """
        Adjust the width of all lines and set the fitted energy resolution
        to the spectrum

        Parameters
        ----------
        xray_lines: list of str or 'all_alpha'
            The Xray lines. If 'all_alpha', fit all using all alpha lines
        ref: list of float
            The centres, before fitting, of the X-ray lines included

        Raises
        ------
        NotImplementedError
            If the signal axis is a non-uniform axis.
        """
        # Test that signal axis is uniform
        if not self.axes_manager[-1].is_uniform:
            raise NotImplementedError(
                "This function is not yet implemented " "for non-uniform axes."
            )

        if xray_lines == "all_alpha":
            xray_lines = [compo.name for compo in self.xray_lines]
        diff = self[xray_lines[0]].centre.value - ref[0]
        offset_old = self.signal.axes_manager[-1].offset
        self.signal.axes_manager[-1].offset -= diff
        offset = self.signal.axes_manager[-1].offset
        _logger.info("Offset changed from  %lf to %lf", offset_old, offset)
        for i, xray_line in enumerate(xray_lines):
            component = self[xray_line]
            component.centre.value = ref[i]

    def calibrate_energy_axis(
        self, calibrate="resolution", xray_lines="all_alpha", **kwargs
    ):
        """
        Calibrate the resolution, the scale or the offset of the energy axis
        by fitting.

        Parameters
        ----------
        calibrate: 'resolution' or 'scale' or 'offset'
            If 'resolution', fits the width of Gaussians place at all x-ray
            lines. The width is given by a model of the detector resolution,
            obtained by extrapolating the `energy_resolution_MnKa` in `metadata`.
            This method will update the value of `energy_resolution_MnKa`.
            If 'scale', calibrate the scale of the energy axis.
            If 'offset', calibrate the offset of the energy axis.
        xray_lines: list of str or 'all_alpha'
            The Xray lines. If 'all_alpha', fit all using all alpha lines
        **kwargs : extra key word arguments
            All extra key word arguments are passed to fit or
            multifit, depending on the value of kind.

        """

        if calibrate == "resolution":
            free = self._twin_xray_lines_width
            fix = self.fix_xray_lines_width
            scale = self._set_energy_resolution
        elif calibrate == "scale":
            free = self._twin_xray_lines_scale
            fix = self.fix_xray_lines_energy
            scale = self._set_energy_scale
        elif calibrate == "offset":
            free = self._twin_xray_lines_offset
            fix = self.fix_xray_lines_energy
            scale = self._set_energy_offset
        ref = free(xray_lines=xray_lines)
        self.fit(**kwargs)
        fix(xray_lines=xray_lines)
        scale(xray_lines=xray_lines, ref=ref)
        self.update_plot()

    def free_sub_xray_lines_weight(self, xray_lines="all", bound=0.01):
        """
        Free the weight of a sub X-ray lines

        Remove the twin on the height of sub-Xray lines (non alpha)

        Parameters
        ----------
        xray_lines : list of str or 'all'
            The Xray lines. If 'all', fit all lines
        bound : float
            Bound the height of the peak to a fraction of
            its height
        """

        def free_twin(component):
            component.A.twin = None
            component.A.free = True
            if component.A.value - bound * component.A.value <= 0:
                component.A.bmin = 1e-10
            else:
                component.A.bmin = component.A.value - bound * component.A.value
            component.A.bmax = component.A.value + bound * component.A.value
            component.A.ext_force_positive = True

        xray_families = [utils_eds._get_xray_lines_family(line) for line in xray_lines]
        for component in self:
            if component.isbackground is False:
                if xray_lines == "all":
                    free_twin(component)
                elif utils_eds._get_xray_lines_family(component.name) in xray_families:
                    free_twin(component)

    def fix_sub_xray_lines_weight(self, xray_lines="all"):
        """
        Fix the weight of a sub X-ray lines to the main X-ray lines

        Establish the twin on the height of sub-Xray lines (non alpha)
        """

        def fix_twin(component):
            component.A.bmin = 0.0
            component.A.bmax = None
            element, line = utils_eds._get_element_and_line(component.name)
            for li in elements_db[element]["Atomic_properties"]["Xray_lines"]:
                if line[0] in li and line != li:
                    xray_sub = element + "_" + li
                    if xray_sub in self:
                        component_sub = self[xray_sub]
                        component_sub.A.bmin = 1e-10
                        component_sub.A.bmax = None
                        weight_line = component_sub.A.value / component.A.value
                        component_sub.A.twin_function_expr = _get_weight(
                            element, li, weight_line
                        )
                        component_sub.A.twin = component.A
                    else:
                        warnings.warn(
                            "The X-ray line expected to be in the "
                            "model was not found"
                        )

        for component in self.xray_lines:
            if xray_lines == "all" or component.name in xray_lines:
                fix_twin(component)
        self.fetch_stored_values()

    def free_xray_lines_energy(self, xray_lines="all", bound=0.001):
        """
        Free the X-ray line energy (shift or centre of the Gaussian)

        Parameters
        ----------
        xray_lines: list of str or 'all'
            The Xray lines. If 'all', fit all lines
        bound: float
            the bound around the actual energy, in keV or eV
        """

        for component in self:
            if component.isbackground is False:
                if xray_lines == "all" or component.name in xray_lines:
                    component.centre.free = True
                    component.centre.bmin = component.centre.value - bound
                    component.centre.bmax = component.centre.value + bound

    def fix_xray_lines_energy(self, xray_lines="all"):
        """
        Fix the X-ray line energy (shift or centre of the Gaussian)

        Parameters
        ----------
        xray_lines: list of str, 'all', or 'all_alpha'
            The Xray lines. If 'all', fit all lines. If 'all_alpha' fit all
            using all alpha lines.
        bound: float
            the bound around the actual energy, in keV or eV
        """
        if xray_lines == "all_alpha":
            xray_lines = [compo.name for compo in self.xray_lines]
        for component in self:
            if component.isbackground is False:
                if xray_lines == "all" or component.name in xray_lines:
                    component.centre.twin = None
                    component.centre.free = False
                    component.centre.bmin = None
                    component.centre.bmax = None

    def free_xray_lines_width(self, xray_lines="all", bound=0.01):
        """
        Free the X-ray line width (sigma of the Gaussian)

        Parameters
        ----------
        xray_lines: list of str or 'all'
            The Xray lines. If 'all', fit all lines
        bound: float
            the bound around the actual energy, in keV or eV
        """

        for component in self:
            if component.isbackground is False:
                if xray_lines == "all" or component.name in xray_lines:
                    component.sigma.free = True
                    component.sigma.bmin = component.sigma.value - bound
                    component.sigma.bmax = component.sigma.value + bound

    def fix_xray_lines_width(self, xray_lines="all"):
        """
        Fix the X-ray line width (sigma of the Gaussian)

        Parameters
        ----------
        xray_lines: list of str, 'all', or 'all_alpha'
            The Xray lines. If 'all', fit all lines. If 'all_alpha' fit all
            using all alpha lines.
        bound: float
            the bound around the actual energy, in keV or eV
        """
        if xray_lines == "all_alpha":
            xray_lines = [compo.name for compo in self.xray_lines]
        for component in self:
            if component.isbackground is False:
                if xray_lines == "all" or component.name in xray_lines:
                    component.sigma.twin = None
                    component.sigma.free = False
                    component.sigma.bmin = None
                    component.sigma.bmax = None

    def calibrate_xray_lines(
        self, calibrate="energy", xray_lines="all", bound=1, kind="single", **kwargs
    ):
        """
        Calibrate individually the X-ray line parameters.

        The X-ray line energy, the weight of the sub-lines and the X-ray line
        width can be calibrated.

        Parameters
        ----------
        calibrate: 'energy' or 'sub_weight' or 'width'
            If 'energy', calibrate the X-ray line energy.
            If 'sub_weight', calibrate the ratio between the main line
            alpha and the other sub-lines of the family
            If 'width', calibrate the X-ray line width.
        xray_lines: list of str or 'all'
            The Xray lines. If 'all', fit all lines
        bounds: float
            for 'energy' and 'width' the bound in energy, in eV
            for 'sub_weight' Bound the height of the peak to fraction of
            its height
        kind : {'single', 'multi'}
            If 'single' fit only the current location. If 'multi'
            use multifit.
        **kwargs : extra key word arguments
            All extra key word arguments are passed to fit or
            multifit, depending on the value of kind.
        """

        if calibrate == "energy":
            bound = (bound / eV2keV) * self.units_factor
            free = self.free_xray_lines_energy
            fix = self.fix_xray_lines_energy
        elif calibrate == "sub_weight":
            free = self.free_sub_xray_lines_weight
            fix = self.fix_sub_xray_lines_weight
        elif calibrate == "width":
            bound = (bound / eV2keV) * self.units_factor
            free = self.free_xray_lines_width
            fix = self.fix_xray_lines_width

        free(xray_lines=xray_lines, bound=bound)
        if kind == "single":
            self.fit(bounded=True, **kwargs)
        elif kind == "multi":
            self.multifit(bounded=True, **kwargs)
        fix(xray_lines=xray_lines)

    def get_lines_intensity(
        self,
        xray_lines=None,
        plot_result=False,
        only_one=True,
        only_lines=("a",),
        **kwargs,
    ):
        """
        Return the fitted intensity of the X-ray lines.

        Return the area under the gaussian corresping to the X-ray lines

        Parameters
        ----------
        xray_lines: {None, list of string}
            If None,
            if `metadata.Sample.elements.xray_lines` contains a
            list of lines use those.
            If `metadata.Sample.elements.xray_lines` is undefined
            or empty but `metadata.Sample.elements` is defined,
            use the same syntax as `add_line` to select a subset of lines
            for the operation.
            Alternatively, provide an iterable containing
            a list of valid X-ray lines symbols.
        plot_result : bool
            If True, plot the calculated line intensities. If the current
            object is a single spectrum it prints the result instead.
        only_one : bool
            If False, use all the lines of each element in the data spectral
            range. If True use only the line at the highest energy
            above an overvoltage of 2 (< beam energy / 2).
        only_lines : {None, list of strings}
            If not None, use only the given lines.
        kwargs
            The extra keyword arguments for plotting. See
            `utils.plot.plot_signals`

        Returns
        -------
        intensities : list
            A list containing the intensities as Signal subclasses.

        Examples
        --------
        >>> m.multifit()
        >>> m.get_lines_intensity(["C_Ka", "Ta_Ma"])
        """
        from hyperspy import utils

        intensities = []

        if xray_lines is None:
            xray_lines = [component.name for component in self.xray_lines]
        else:
            xray_lines = self.signal._parse_xray_lines(xray_lines, only_one, only_lines)
            xray_lines = list(filter(lambda x: x in [a.name for a in self], xray_lines))
        if len(xray_lines) == 0:
            raise ValueError("These X-ray lines are not part of the model.")

        for xray_line in xray_lines:
            element, line = utils_eds._get_element_and_line(xray_line)
            line_energy = self.signal._get_line_energy(xray_line)
            data_res = self[xray_line].A.map["values"]
            if self.axes_manager.navigation_dimension == 0:
                data_res = data_res[0]
            img = self.signal.isig[0:1].integrate1D(-1)
            img.data = data_res
            img.metadata.General.title = "Intensity of %s at %.2f %s from %s" % (
                xray_line,
                line_energy,
                self.signal.axes_manager.signal_axes[0].units,
                self.signal.metadata.General.title,
            )
            img = img.transpose(signal_axes=[])
            if plot_result and img.axes_manager.signal_dimension == 0:
                print(
                    "%s at %s %s : Intensity = %.2f"
                    % (
                        xray_line,
                        line_energy,
                        self.signal.axes_manager.signal_axes[0].units,
                        img.data,
                    )
                )
            img.metadata.set_item("Sample.elements", ([element]))
            img.metadata.set_item("Sample.xray_lines", ([xray_line]))
            intensities.append(img)
        if plot_result and img.axes_manager.signal_dimension != 0:
            utils.plot.plot_signals(intensities, **kwargs)
        return intensities

    def remove(self, thing):
        thing = self._get_component(thing)
        if not np.iterable(thing):
            thing = [
                thing,
            ]
        for comp in thing:
            if comp in self.xray_lines:
                self.xray_lines.remove(comp)
            elif comp in self.family_lines:
                self.family_lines.remove(comp)
            elif comp in self.background_components:
                self.background_components.remove(comp)
        super().remove(thing)



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.
from copy import deepcopy

import numpy as np

import exspy
from exspy.signal_tools import EdgesRange


class Owner:
    """for use in Test_EdgesRange"""

    def __init__(self, edge):
        self.description = edge


class Test_EdgesRange:
    def setup_method(self, method):
        s = exspy.data.EELS_MnFe()
        er = EdgesRange(s)
        self.signal = s
        self.er = er

    def test_init(self):
        edges_all = np.array(
            [
                "Ag_M2",
                "Ra_N5",
                "Fr_N4",
                "Cr_L2",
                "Cd_M3",
                "Te_M4",
                "I_M5",
                "Fr_N5",
                "Cr_L3",
                "Te_M5",
                "V_L1",
                "Ag_M3",
                "I_M4",
                "Rn_N4",
                "Ti_L1",
                "Ra_N4",
                "Mn_L3",
                "Pd_M2",
                "Cd_M2",
                "Mn_L2",
                "Tc_M1",
                "Sb_M4",
                "In_M3",
                "At_N5",
                "O_K",
                "Pd_M3",
                "Sb_M5",
                "Xe_M5",
                "Ac_N4",
                "Rh_M2",
                "V_L2",
                "F_K",
                "Xe_M4",
                "V_L3",
                "Cr_L1",
                "Sc_L1",
                "In_M2",
                "Rh_M3",
                "Sn_M4",
                "Pa_N5",
                "Fe_L3",
                "Sn_M3",
                "Sn_M5",
                "Ru_M2",
                "Fe_L2",
                "Cs_M5",
                "Ti_L2",
                "Ru_M3",
                "Cs_M4",
                "At_N3",
                "Pa_N4",
                "Ti_L3",
                "In_M4",
                "Pu_N6",
                "Tc_M2",
                "Sn_M2",
                "In_M5",
                "Ca_L1",
                "Sb_M3",
                "Pu_N7",
                "Rn_N3",
                "Mn_L1",
                "Np_N5",
                "Tc_M3",
                "Co_L3",
                "Ba_M5",
                "Np_N6",
                "Cd_M4",
                "Mo_M2",
                "Sc_L2",
                "Co_L2",
                "Cd_M5",
                "Np_N7",
                "Ba_M4",
                "Sc_L3",
                "N_K",
            ]
        )
        energy_all = np.array(
            [
                602.0,
                603.0,
                603.0,
                584.0,
                616.0,
                582.0,
                620.0,
                577.0,
                575.0,
                572.0,
                628.0,
                571.0,
                631.0,
                567.0,
                564.0,
                636.0,
                640.0,
                559.0,
                651.0,
                651.0,
                544.0,
                537.0,
                664.0,
                533.0,
                532.0,
                531.0,
                528.0,
                672.0,
                675.0,
                521.0,
                521.0,
                685.0,
                685.0,
                513.0,
                695.0,
                500.0,
                702.0,
                496.0,
                494.0,
                708.0,
                708.0,
                714.0,
                485.0,
                483.0,
                721.0,
                726.0,
                462.0,
                461.0,
                740.0,
                740.0,
                743.0,
                456.0,
                451.0,
                446.0,
                445.0,
                756.0,
                443.0,
                438.0,
                766.0,
                432.0,
                768.0,
                769.0,
                770.0,
                425.0,
                779.0,
                781.0,
                415.0,
                411.0,
                410.0,
                407.0,
                794.0,
                404.0,
                404.0,
                796.0,
                402.0,
                401.0,
            ]
        )
        relevance_all = np.array(
            [
                "Minor",
                "Minor",
                "Minor",
                "Major",
                "Minor",
                "Major",
                "Major",
                "Minor",
                "Major",
                "Major",
                "Minor",
                "Minor",
                "Major",
                "Minor",
                "Minor",
                "Minor",
                "Major",
                "Minor",
                "Minor",
                "Major",
                "Minor",
                "Major",
                "Minor",
                "Minor",
                "Major",
                "Minor",
                "Major",
                "Major",
                "Minor",
                "Minor",
                "Major",
                "Major",
                "Major",
                "Major",
                "Minor",
                "Minor",
                "Minor",
                "Minor",
                "Major",
                "Minor",
                "Major",
                "Minor",
                "Major",
                "Minor",
                "Major",
                "Major",
                "Major",
                "Minor",
                "Major",
                "Minor",
                "Minor",
                "Major",
                "Major",
                "Major",
                "Minor",
                "Minor",
                "Major",
                "Minor",
                "Minor",
                "Major",
                "Minor",
                "Minor",
                "Minor",
                "Minor",
                "Major",
                "Major",
                "Major",
                "Major",
                "Minor",
                "Major",
                "Major",
                "Major",
                "Major",
                "Major",
                "Major",
                "Major",
            ]
        )
        description_all = np.array(
            [
                "Delayed maximum",
                "",
                "",
                "Sharp peak. Delayed maximum",
                "Delayed maximum",
                "Delayed maximum",
                "Delayed maximum",
                "",
                "Sharp peak. Delayed maximum",
                "Delayed maximum",
                "Abrupt onset",
                "Delayed maximum",
                "Delayed maximum",
                "",
                "Abrupt onset",
                "",
                "Sharp peak. Delayed maximum",
                "",
                "",
                "Sharp peak. Delayed maximum",
                "Abrupt onset",
                "Delayed maximum",
                "Delayed maximum",
                "",
                "Abrupt onset",
                "",
                "Delayed maximum",
                "Delayed maximum",
                "",
                "Sharp peak",
                "Sharp peak. Delayed maximum",
                "Abrupt onset",
                "Delayed maximum",
                "Sharp peak. Delayed maximum",
                "Abrupt onset",
                "Abrupt onset",
                "",
                "Sharp peak",
                "Delayed maximum",
                "",
                "Sharp peak. Delayed maximum",
                "Delayed maximum",
                "Delayed maximum",
                "Sharp peak",
                "Sharp peak. Delayed maximum",
                "Sharp peak. Delayed maximum",
                "Sharp peak. Delayed maximum",
                "Sharp peak",
                "Sharp peak. Delayed maximum",
                "",
                "",
                "Sharp peak. Delayed maximum",
                "Delayed maximum",
                "",
                "Sharp peak. Delayed maximum",
                "",
                "Delayed maximum",
                "Abrupt onset",
                "Delayed maximum",
                "",
                "",
                "Abrupt onset",
                "",
                "Sharp peak. Delayed maximum",
                "Sharp peak. Delayed maximum",
                "Sharp peak. Delayed maximum",
                "",
                "Delayed maximum",
                "Sharp peak",
                "Sharp peak. Delayed maximum",
                "Sharp peak. Delayed maximum",
                "Delayed maximum",
                "",
                "Sharp peak. Delayed maximum",
                "Sharp peak. Delayed maximum",
                "Abrupt onset",
            ]
        )

        assert np.array_equal(self.er.edge_all, edges_all)
        assert np.array_equal(self.er.energy_all, energy_all)
        assert np.array_equal(self.er.relevance_all, relevance_all)
        assert np.array_equal(self.er.description_all, description_all)

    def test_selected_span_selector(self):
        # self.er.span_selector.extents = (500, 550)
        self.er.ss_left_value = 500
        self.er.ss_right_value = 550

        edges, energy, relevance, description = self.er.update_table()
        assert set(edges) == set(
            (
                "Tc_M1",
                "Sb_M4",
                "At_N5",
                "O_K",
                "Pd_M3",
                "Sb_M5",
                "Rh_M2",
                "V_L2",
                "V_L3",
                "Sc_L1",
            )
        )
        assert set(energy) == set(
            (544.0, 537.0, 533.0, 532.0, 531.0, 528.0, 521.0, 521.0, 513.0, 500.0)
        )
        assert set(relevance) == set(
            (
                "Minor",
                "Major",
                "Minor",
                "Major",
                "Minor",
                "Major",
                "Minor",
                "Major",
                "Major",
                "Minor",
            )
        )
        assert set(description) == set(
            (
                "Abrupt onset",
                "Delayed maximum",
                "",
                "Abrupt onset",
                "",
                "Delayed maximum",
                "Sharp peak",
                "Sharp peak. Delayed maximum",
                "Sharp peak. Delayed maximum",
                "Abrupt onset",
            )
        )

    def test_none_span_selector(self):
        self.er.span_selector = None

        edges, energy, relevance, description = self.er.update_table()

        assert len(edges) == 0
        assert len(energy) == 0
        assert len(relevance) == 0
        assert len(description) == 0

    def test_complementary_edge(self):
        self.signal.plot(plot_edges=["V_L2"])
        er = EdgesRange(self.signal)
        er.ss_right_value = 550
        er.ss_left_value = 500
        _ = er.update_table()

        assert er.active_edges == ["V_L2"]
        assert er.active_complementary_edges == ["V_L3", "V_L1"]

    def test_off_complementary_edge(self):
        self.signal.plot(plot_edges=["V_L2"])
        er = EdgesRange(self.signal)
        er.complementary = False
        er.ss_right_value = 550
        er.ss_left_value = 500
        _ = er.update_table()

        assert er.active_edges == ["V_L2"]
        assert len(er.active_complementary_edges) == 0

    def test_keep_valid_edge(self):
        self.signal.plot(plot_edges=["V_L2"])
        er = EdgesRange(self.signal)
        er.ss_right_value = 550
        er.ss_left_value = 500
        _ = er.update_table()

        er.ss_right_value = 650
        er.ss_left_value = 600
        _ = er.update_table()

        assert er.active_edges == ["V_L1"]
        assert er.active_complementary_edges == ["V_L2", "V_L3"]

    def test_remove_out_of_range_edge(self):
        self.signal.plot(plot_edges=["V_L2"])
        er = EdgesRange(self.signal)
        er.ss_right_value = 550
        er.ss_left_value = 500
        _ = er.update_table()

        er.ss_right_value = 750
        er.ss_left_value = 700
        _ = er.update_table()

        assert len(er.active_edges) == 0
        assert len(er.active_complementary_edges) == 0

    def test_select_edge_by_button(self):
        self.er.ss_left_value = 500
        self.er.ss_right_value = 550
        _ = self.er.update_table()

        on_V_L2 = {"owner": Owner("V_L2"), "new": True}
        self.er.update_active_edge(on_V_L2)

        assert self.er.active_edges == ["V_L2"]
        assert self.er.active_complementary_edges == ["V_L3", "V_L1"]

        off_V_L2 = {"owner": Owner("V_L2"), "new": False}
        self.er.update_active_edge(off_V_L2)

        assert len(self.er.active_edges) == 0
        assert len(self.er.active_complementary_edges) == 0

    def test_remove_all_edge_markers(self):
        self.signal.plot(plot_edges=["V_L2"])
        er = EdgesRange(self.signal)
        er.ss_right_value = 550
        er.ss_left_value = 500
        _ = er.update_table()

        er._clear_markers()

        assert len(er.active_edges) == 0
        assert len(er.active_complementary_edges) == 0

    def test_on_figure_changed(self):
        self.signal.plot(plot_edges=["V_L2"])
        er = EdgesRange(self.signal)
        er.ss_right_value = 550
        er.ss_left_value = 500
        _ = er.update_table()

        segments = deepcopy(self.signal._edge_markers["lines"].get_current_kwargs())
        scaled = self.signal._edge_markers["lines"]._scale_kwarg(segments, "segments")
        self.signal._plot.pointer.indices = (9,)
        assert self.signal.axes_manager.navigation_axes[0].index == 9
        segments2 = deepcopy(self.signal._edge_markers["lines"].get_current_kwargs())
        scaled2 = self.signal._edge_markers["lines"]._scale_kwarg(segments2, "segments")
        assert not np.array_equal(scaled["segments"], scaled2["segments"])



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

from collections.abc import Iterable
import numpy as np
import numbers
import copy

from exspy.misc.elements import elements as elements_db
from exspy.misc.eds.ffast_mac import ffast_mac_db as ffast_mac
from exspy.misc.eds import utils as utils_eds
from hyperspy.misc.utils import stack


__all__ = [
    "atomic_to_weight",
    "weight_to_atomic",
    "density_of_mixture",
    "mass_absorption_coefficient",
    "mass_absorption_mixture",
]


def __dir__():
    return sorted(__all__)


def _weight_to_atomic(weight_percent, elements):
    """Convert weight percent (wt%) to atomic percent (at.%).

    Parameters
    ----------
    weight_percent: array of float
        The weight fractions (composition) of the sample.
    elements: list of str
        A list of element abbreviations, e.g. ['Al','Zn']

    Returns
    -------
    atomic_percent : array of float
        Composition in atomic percent.

    Calculate the atomic percent of modern bronze given its weight percent:
    >>> hs.material.weight_to_atomic((88, 12), ("Cu", "Sn"))
    array([ 93.19698614,   6.80301386])

    """
    if len(elements) != len(weight_percent):
        raise ValueError(
            "The number of elements must match the size of the first axis"
            "of weight_percent."
        )
    atomic_weights = np.array(
        [
            elements_db[element]["General_properties"]["atomic_weight"]
            for element in elements
        ]
    )
    atomic_percent = np.array(list(map(np.divide, weight_percent, atomic_weights)))
    sum_weight = atomic_percent.sum(axis=0) / 100.0
    for i, el in enumerate(elements):
        atomic_percent[i] /= sum_weight
        atomic_percent[i] = np.where(sum_weight == 0.0, 0.0, atomic_percent[i])
    return atomic_percent


def weight_to_atomic(weight_percent, elements="auto"):
    """Convert weight percent (wt%) to atomic percent (at.%).

    Parameters
    ----------
    weight_percent: list of float or list of signals
        The weight fractions (composition) of the sample.
    elements: list of str
        A list of element abbreviations, e.g. ['Al','Zn']. If elements is
        'auto', take the elements in en each signal metadata of th
        weight_percent list.

    Returns
    -------
    atomic_percent : as weight_percent
        Composition in atomic percent.

    Examples
    --------
    Calculate the atomic percent of modern bronze given its weight percent:
    >>> hs.material.weight_to_atomic((88, 12), ("Cu", "Sn"))
    array([ 93.19698614,   6.80301386])

    """
    from hyperspy.signals import BaseSignal

    elements = _elements_auto(weight_percent, elements)

    if isinstance(weight_percent[0], BaseSignal):
        atomic_percent = stack(weight_percent)
        atomic_percent.data = _weight_to_atomic(atomic_percent.data, elements)
        atomic_percent.data = np.nan_to_num(atomic_percent.data)
        atomic_percent = atomic_percent.split()
        for i, el in enumerate(elements):
            atomic_percent[i].metadata.General.title = "atomic percent of " + el
        return atomic_percent
    else:
        return _weight_to_atomic(weight_percent, elements)


def _atomic_to_weight(atomic_percent, elements):
    """Convert atomic percent to weight percent.

    Parameters
    ----------
    atomic_percent: array
        The atomic fractions (composition) of the sample.
    elements: list of str
        A list of element abbreviations, e.g. ['Al','Zn']

    Returns
    -------
    weight_percent : array of float
        composition in weight percent.

    Examples
    --------
    Calculate the weight percent of modern bronze given its atomic percent:
    >>> hs.material.atomic_to_weight([93.2, 6.8], ("Cu", "Sn"))
    array([ 88.00501989,  11.99498011])

    """
    if len(elements) != len(atomic_percent):
        raise ValueError(
            "The number of elements must match the size of the first axis"
            "of atomic_percent."
        )
    atomic_weights = np.array(
        [
            elements_db[element]["General_properties"]["atomic_weight"]
            for element in elements
        ]
    )
    weight_percent = np.array(list(map(np.multiply, atomic_percent, atomic_weights)))
    sum_atomic = weight_percent.sum(axis=0) / 100.0
    for i, el in enumerate(elements):
        weight_percent[i] /= sum_atomic
        weight_percent[i] = np.where(sum_atomic == 0.0, 0.0, weight_percent[i])
    return weight_percent


def atomic_to_weight(atomic_percent, elements="auto"):
    """Convert atomic percent to weight percent.

    Parameters
    ----------
    atomic_percent: list of float or list of signals
        The atomic fractions (composition) of the sample.
    elements: list of str
        A list of element abbreviations, e.g. ['Al','Zn']. If elements is
        'auto', take the elements in en each signal metadata of the
        atomic_percent list.

    Returns
    -------
    weight_percent : as atomic_percent
        composition in weight percent.

    Examples
    --------
    Calculate the weight percent of modern bronze given its atomic percent:
    >>> hs.material.atomic_to_weight([93.2, 6.8], ("Cu", "Sn"))
    array([ 88.00501989,  11.99498011])

    """
    from hyperspy.signals import BaseSignal

    elements = _elements_auto(atomic_percent, elements)
    if isinstance(atomic_percent[0], BaseSignal):
        weight_percent = stack(atomic_percent, show_progressbar=False)
        weight_percent.data = _atomic_to_weight(weight_percent.data, elements)
        weight_percent = weight_percent.split()
        for i, el in enumerate(elements):
            atomic_percent[i].metadata.General.title = "weight percent of " + el
        return weight_percent
    else:
        return _atomic_to_weight(atomic_percent, elements)


def _density_of_mixture(weight_percent, elements, mean="harmonic"):
    """Calculate the density a mixture of elements.

    The density of the elements is retrieved from an internal database. The
    calculation is only valid if there is no interaction between the
    components.

    Parameters
    ----------
    weight_percent: array
        A list of weight percent for the different elements. If the total
        is not equal to 100, each weight percent is divided by the sum
        of the list (normalization).
    elements: list of str
        A list of element symbols, e.g. ['Al', 'Zn']
    mean: 'harmonic' or 'weighted'
        The type of mean use to estimate the density

    Returns
    -------
    density: The density in g/cm3.

    Examples
    --------
    Calculate the density of modern bronze given its weight percent:
    >>> hs.material.density_of_mixture([88, 12],['Cu', 'Sn'])
    8.6903187973131466

    """
    if len(elements) != len(weight_percent):
        raise ValueError(
            "The number of elements must match the size of the first axis"
            "of weight_percent."
        )
    densities = np.array(
        [
            elements_db[element]["Physical_properties"]["density (g/cm^3)"]
            for element in elements
        ]
    )
    sum_densities = np.zeros_like(weight_percent, dtype="float")
    try:
        if mean == "harmonic":
            for i, weight in enumerate(weight_percent):
                sum_densities[i] = weight / densities[i]
            sum_densities = sum_densities.sum(axis=0)
            density = np.sum(weight_percent, axis=0) / sum_densities
            return np.where(sum_densities == 0.0, 0.0, density)
        elif mean == "weighted":
            for i, weight in enumerate(weight_percent):
                sum_densities[i] = weight * densities[i]
            sum_densities = sum_densities.sum(axis=0)
            sum_weight = np.sum(weight_percent, axis=0)
            density = sum_densities / sum_weight
            return np.where(sum_weight == 0.0, 0.0, density)
    except TypeError:
        raise ValueError(
            "The density of one of the elements is unknown (Probably At or Fr)."
        )


def density_of_mixture(weight_percent, elements="auto", mean="harmonic"):
    """Calculate the density of a mixture of elements.

    The density of the elements is retrieved from an internal database. The
    calculation is only valid if there is no interaction between the
    components.

    Parameters
    ----------
    weight_percent: list of float or list of signals
        A list of weight percent for the different elements. If the total
        is not equal to 100, each weight percent is divided by the sum
        of the list (normalization).
    elements: list of str
        A list of element symbols, e.g. ['Al', 'Zn']. If elements is 'auto',
        take the elements in en each signal metadata of the weight_percent
        list.
    mean: 'harmonic' or 'weighted'
        The type of mean use to estimate the density

    Returns
    -------
    density: The density in g/cm3.

    Examples
    --------
    Calculate the density of modern bronze given its weight percent:
    >>> hs.material.density_of_mixture([88, 12],['Cu', 'Sn'])
    8.6903187973131466

    """
    from hyperspy.signals import BaseSignal

    elements = _elements_auto(weight_percent, elements)
    if isinstance(weight_percent[0], BaseSignal):
        density = weight_percent[0]._deepcopy_with_new_data(
            _density_of_mixture(stack(weight_percent).data, elements, mean=mean)
        )
        return density
    else:
        return _density_of_mixture(weight_percent, elements, mean=mean)


def mass_absorption_coefficient(element, energies):
    """
    Mass absorption coefficient (mu/rho) of a X-ray absorbed in a pure
    material.

    The mass absorption is retrieved from the database of Chantler2005

    Parameters
    ----------
    element: str
        The element symbol of the absorber, e.g. 'Al'.
    energies: float or list of float or str or list of str
        The energy or energies of the X-ray in keV, or the name of the X-rays,
        e.g. 'Al_Ka'.

    Return
    ------
    mass absorption coefficient(s) in cm^2/g

    Examples
    --------
    >>> hs.material.mass_absorption_coefficient(
    >>>     element='Al', energies=['C_Ka','Al_Ka'])
    array([ 26330.38933818,    372.02616732])

    See also
    --------
    :py:func:`~hs.material.mass_absorption_mixture`

    Note
    ----
    See https://dx.doi.org/10.18434/T4HS32
    Chantler, C.T., Olsen, K., Dragoset, R.A., Kishore, A.R., Kotochigova,
    S.A., and Zucker, D.S. (2005), X-Ray Form Factor, Attenuation and
    Scattering Tables (version 2.1).
    """
    energies_db = np.array(ffast_mac[element].energies_keV)
    macs = np.array(ffast_mac[element].mass_absorption_coefficient_cm2g)
    energies = copy.copy(energies)
    if isinstance(energies, str):
        energies = utils_eds._get_energy_xray_line(energies)
    elif isinstance(energies, Iterable):
        for i, energy in enumerate(energies):
            if isinstance(energy, str):
                energies[i] = utils_eds._get_energy_xray_line(energy)
    index = np.searchsorted(energies_db, energies)
    mac_res = np.exp(
        np.log(macs[index - 1])
        + np.log(macs[index] / macs[index - 1])
        * (
            np.log(energies / energies_db[index - 1])
            / np.log(energies_db[index] / energies_db[index - 1])
        )
    )
    return np.nan_to_num(mac_res)


def _mass_absorption_mixture(weight_percent, elements, energies):
    """Calculate the mass absorption coefficient for X-ray absorbed in a
    mixture of elements.

    The mass absorption coefficient is calculated as a weighted mean of the
    weight percent and is retrieved from the database of Chantler2005.

    Parameters
    ----------
    weight_percent: np.array
        The composition of the absorber(s) in weight percent. The first
        dimension of the matrix corresponds to the elements.
    elements: list of str
        The list of element symbol of the absorber, e.g. ['Al','Zn'].
    energies: float or list of float or str or list of str
        The energy or energies of the X-ray in keV, or the name of the X-rays,
        e.g. 'Al_Ka'.

    Examples
    --------
    >>> hs.material.mass_absorption_mixture(
    >>>     elements=['Al','Zn'], weight_percent=[50,50], energies='Al_Ka')
    2587.4161643905127

    Return
    ------
    float or array of float
    mass absorption coefficient(s) in cm^2/g

    See also
    --------
    :py:func:`~hs.material.mass_absorption`

    Note
    ----
    See https://dx.doi.org/10.18434/T4HS32
    Chantler, C.T., Olsen, K., Dragoset, R.A., Kishore, A.R., Kotochigova,
    S.A., and Zucker, D.S. (2005), X-Ray Form Factor, Attenuation and
    Scattering Tables (version 2.1).
    """
    if len(elements) != len(weight_percent):
        raise ValueError("Elements and weight_fraction should have the same length")
    if isinstance(weight_percent[0], Iterable):
        weight_fraction = np.array(weight_percent)
        weight_fraction /= np.sum(weight_fraction, 0)
        mac_res = np.zeros([len(energies)] + list(weight_fraction.shape[1:]))
        for element, weight in zip(elements, weight_fraction):
            mac_re = mass_absorption_coefficient(element, energies)
            mac_res += np.array([weight * ma for ma in mac_re])
        return mac_res
    else:
        mac_res = np.array(
            [mass_absorption_coefficient(el, energies) for el in elements]
        )
        mac_res = np.dot(weight_percent, mac_res) / np.sum(weight_percent, 0)
        return mac_res


def mass_absorption_mixture(weight_percent, elements="auto", energies="auto"):
    """Calculate the mass absorption coefficient for X-ray absorbed in a
    mixture of elements.

    The mass absorption coefficient is calculated as a weighted mean of the
    weight percent and is retrieved from the database of Chantler2005.

    Parameters
    ----------
    weight_percent: list of float or list of signals
        The composition of the absorber(s) in weight percent. The first
        dimension of the matrix corresponds to the elements.
    elements: list of str or 'auto'
        The list of element symbol of the absorber, e.g. ['Al','Zn']. If
        elements is 'auto', take the elements in each signal metadata of the
        weight_percent list.
    energies: list of float or list of str or 'auto'
        The energy or energies of the X-ray in keV, or the name of the X-rays,
        e.g. 'Al_Ka'. If 'auto', take the lines in each signal metadata of the
        weight_percent list.

    Examples
    --------
    >>> hs.material.mass_absorption_mixture(
    >>>     elements=['Al','Zn'], weight_percent=[50,50], energies='Al_Ka')
    2587.41616439

    Return
    ------
    float or array of float
    mass absorption coefficient(s) in cm^2/g

    See also
    --------
    :py:func:`~hs.material.mass_absorption_coefficient`

    Note
    ----
    See https://dx.doi.org/10.18434/T4HS32
    Chantler, C.T., Olsen, K., Dragoset, R.A., Kishore, A.R., Kotochigova,
    S.A., and Zucker, D.S. (2005), X-Ray Form Factor, Attenuation and
    Scattering Tables (version 2.1).

    """
    from hyperspy.signals import BaseSignal

    elements = _elements_auto(weight_percent, elements)
    energies = _lines_auto(weight_percent, energies)
    if isinstance(weight_percent[0], BaseSignal):
        weight_per = np.array([wt.data for wt in weight_percent])
        mac_res = stack(
            [weight_percent[0].deepcopy()] * len(energies), show_progressbar=False
        )
        mac_res.data = _mass_absorption_mixture(weight_per, elements, energies)
        mac_res = mac_res.split()
        for i, energy in enumerate(energies):
            mac_res[i].metadata.set_item("Sample.xray_lines", ([energy]))
            mac_res[i].metadata.General.set_item(
                "title",
                "Absoprtion coeff of"
                " %s in %s" % (energy, mac_res[i].metadata.General.title),
            )
            if mac_res[i].metadata.has_item("Sample.elements"):
                del mac_res[i].metadata.Sample.elements
        return mac_res
    else:
        return _mass_absorption_mixture(weight_percent, elements, energies)


def _lines_auto(composition, xray_lines):
    if isinstance(composition[0], numbers.Number):
        if isinstance(xray_lines, str):
            if xray_lines == "auto":
                raise ValueError("The X-ray lines needs to be provided.")
    else:
        if isinstance(xray_lines, str):
            if xray_lines == "auto":
                xray_lines = []
                for compo in composition:
                    if len(compo.metadata.Sample.xray_lines) > 1:
                        raise ValueError(
                            "The signal %s contains more than one X-ray lines "
                            "but this function requires only one X-ray lines "
                            "per signal." % compo.metadata.General.title
                        )
                    else:
                        xray_lines.append(compo.metadata.Sample.xray_lines[0])
    return xray_lines


def _elements_auto(composition, elements):
    if isinstance(composition[0], numbers.Number):
        if isinstance(elements, str):
            if elements == "auto":
                raise ValueError("The elements needs to be provided.")
    else:
        if isinstance(elements, str):
            if elements == "auto":
                elements = []
                for compo in composition:
                    if len(compo.metadata.Sample.elements) > 1:
                        raise ValueError(
                            "The signal %s contains more than one element "
                            "but this function requires only one element "
                            "per signal." % compo.metadata.General.title
                        )
                    else:
                        elements.append(compo.metadata.Sample.elements[0])
    return elements



from exspy.misc.eels.hydrogenic_gos import HydrogenicGOS
from exspy.misc.eels.gosh_gos import GoshGOS
from exspy.misc.eels.hartree_slater_gos import HartreeSlaterGOS

__all__ = ["HydrogenicGOS", "GoshGOS", "HartreeSlaterGOS"]



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.


import numpy as np
import math
from scipy import constants
from hyperspy.misc.utils import stack
from exspy.misc.elements import elements as elements_db
from functools import reduce


eV2keV = 1000.0
sigma2fwhm = 2 * math.sqrt(2 * math.log(2))


_ABSORPTION_CORRECTION_DOCSTRING = """absorption_correction : numpy.ndarray or None
        If None (default), absorption correction is ignored, otherwise, the
        array must contain values between 0 and 1 to correct the intensities
        based on estimated absorption.
"""


def _get_element_and_line(xray_line):
    """
    Returns the element name and line character for a particular X-ray line as
    a tuple.

    By example, if xray_line = 'Mn_Ka' this function returns ('Mn', 'Ka')
    """
    lim = xray_line.find("_")
    if lim == -1:
        raise ValueError(f"Invalid xray-line: {xray_line}")
    return xray_line[:lim], xray_line[lim + 1 :]


def _get_energy_xray_line(xray_line):
    """
    Returns the energy (in keV) associated with a given X-ray line.

    By example, if xray_line = 'Mn_Ka' this function returns 5.8987
    """
    element, line = _get_element_and_line(xray_line)
    return elements_db[element]["Atomic_properties"]["Xray_lines"][line]["energy (keV)"]


def _get_xray_lines_family(xray_line):
    """
    Returns the family to which a particular X-ray line belongs.

    By example, if xray_line = 'Mn_Ka' this function returns 'Mn_K'
    """
    return xray_line[: xray_line.find("_") + 2]


def _parse_only_lines(only_lines):
    if isinstance(only_lines, str):
        pass
    elif hasattr(only_lines, "__iter__"):
        if any(isinstance(line, str) is False for line in only_lines):
            return only_lines
    else:
        return only_lines
    only_lines = list(only_lines)
    for only_line in only_lines:
        if only_line == "a":
            only_lines.extend(["Ka", "La", "Ma"])
        elif only_line == "b":
            only_lines.extend(["Kb", "Lb1", "Mb"])
    return only_lines


def get_xray_lines_near_energy(energy, width=0.2, only_lines=None):
    """Find xray lines near a specific energy, more specifically all xray lines
    that satisfy only_lines and are within the given energy window width around
    the passed energy.

    Parameters
    ----------
    energy : float
        Energy to search near in keV
    width : float
        Window width in keV around energy in which to find nearby energies,
        i.e. a value of 0.2 keV (the default) means to search +/- 0.1 keV.
    only_lines :
        If not None, only the given lines will be added (eg. ('a','Kb')).

    Returns
    -------
    List of xray-lines sorted by energy difference to given energy.
    """
    only_lines = _parse_only_lines(only_lines)
    valid_lines = []
    E_min, E_max = energy - width / 2.0, energy + width / 2.0
    for element, el_props in elements_db.items():
        # Not all elements in the DB have the keys, so catch KeyErrors
        try:
            lines = el_props["Atomic_properties"]["Xray_lines"]
        except KeyError:
            continue
        for line, l_props in lines.items():
            if only_lines and line not in only_lines:
                continue
            line_energy = l_props["energy (keV)"]
            if E_min <= line_energy <= E_max:
                # Store line in Element_Line format, and energy difference
                valid_lines.append((element + "_" + line, abs(line_energy - energy)))
    # Sort by energy difference, but return only the line names
    return [line for line, _ in sorted(valid_lines, key=lambda x: x[1])]


def get_FWHM_at_Energy(energy_resolution_MnKa, E):
    """Calculates an approximate FWHM, accounting for peak broadening due to the
    detector, for a peak at energy E given a known width at a reference energy.

    The factor 2.5 is a constant derived by Fiori & Newbury as references
    below.

    Parameters
    ----------
    energy_resolution_MnKa : float
        Energy resolution of Mn Ka in eV
    E : float
        Energy of the peak in keV

    Returns
    -------
    float : FWHM of the peak in keV

    Notes
    -----
    This method implements the equation derived by Fiori and Newbury as is
    documented in the following:

        Fiori, C. E., and Newbury, D. E. (1978). In SEM/1978/I, SEM, Inc.,
        AMF O'Hare, Illinois, p. 401.

        Goldstein et al. (2003). "Scanning Electron Microscopy & X-ray
        Microanalysis", Plenum, third edition, p 315.

    """
    FWHM_ref = energy_resolution_MnKa
    E_ref = _get_energy_xray_line("Mn_Ka")

    FWHM_e = 2.5 * (E - E_ref) * eV2keV + FWHM_ref * FWHM_ref

    return math.sqrt(FWHM_e) / 1000.0  # In mrad


def xray_range(xray_line, beam_energy, density="auto"):
    """Return the maximum range of X-ray generation according to the
    Anderson-Hasler parameterization.

    Parameters
    ----------
    xray_line: str
        The X-ray line, e.g. 'Al_Ka'
    beam_energy: float
        The energy of the beam in kV.
    density: {float, 'auto'}
        The density of the material in g/cm3. If 'auto', the density
        of the pure element is used.

    Returns
    -------
    X-ray range in micrometer.

    Examples
    --------
    >>> # X-ray range of Cu Ka in pure Copper at 30 kV in micron
    >>> hs.eds.xray_range('Cu_Ka', 30.)
    1.9361716759499248

    >>> # X-ray range of Cu Ka in pure Carbon at 30kV in micron
    >>> hs.eds.xray_range('Cu_Ka', 30., hs.material.elements.C.
    >>>                      Physical_properties.density_gcm3)
    7.6418811280855454

    Notes
    -----
    From Anderson, C.A. and M.F. Hasler (1966). In proceedings of the
    4th international conference on X-ray optics and microanalysis.

    See also the textbook of Goldstein et al., Plenum publisher,
    third edition p 286

    """

    element, line = _get_element_and_line(xray_line)
    if density == "auto":
        density = elements_db[element]["Physical_properties"]["density (g/cm^3)"]
    Xray_energy = _get_energy_xray_line(xray_line)
    # Note: magic numbers here are from Andersen-Hasler parameterization. See
    # docstring for associated references.
    return 0.064 / density * (np.power(beam_energy, 1.68) - np.power(Xray_energy, 1.68))


def electron_range(element, beam_energy, density="auto", tilt=0):
    """Returns the maximum electron range for a pure bulk material according to
    the Kanaya-Okayama parameterziation.

    Parameters
    ----------
    element: str
        The element symbol, e.g. 'Al'.
    beam_energy: float
        The energy of the beam in keV.
    density: {float, 'auto'}
        The density of the material in g/cm3. If 'auto', the density of
        the pure element is used.
    tilt: float.
        The tilt of the sample in degrees.

    Returns
    -------
    Electron range in micrometers.

    Examples
    --------
    >>> # Electron range in pure Copper at 30 kV in micron
    >>> hs.eds.electron_range('Cu', 30.)
    2.8766744984001607

    Notes
    -----
    From Kanaya, K. and S. Okayama (1972). J. Phys. D. Appl. Phys. 5, p43

    See also the textbook of Goldstein et al., Plenum publisher,
    third edition p 72.

    """

    if density == "auto":
        density = elements_db[element]["Physical_properties"]["density (g/cm^3)"]
    Z = elements_db[element]["General_properties"]["Z"]
    A = elements_db[element]["General_properties"]["atomic_weight"]
    # Note: magic numbers here are from Kanaya-Okayama parameterization. See
    # docstring for associated references.
    return (
        0.0276
        * A
        / np.power(Z, 0.89)
        / density
        * np.power(beam_energy, 1.67)
        * math.cos(math.radians(tilt))
    )


def take_off_angle(tilt_stage, azimuth_angle, elevation_angle, beta_tilt=0.0):
    """Calculate the take-off-angle (TOA).

    TOA is the angle with which the X-rays leave the surface towards
    the detector.

    Parameters
    ----------
    alpha_tilt: float
        The alpha-tilt of the stage in degrees. The sample is facing the detector
        when positively tilted.
    azimuth_angle: float
        The azimuth of the detector in degrees. 0 is perpendicular to the alpha
        tilt axis.
    elevation_angle: float
        The elevation of the detector in degrees.
    beta_tilt: float
        The beta-tilt of the stage in degrees. The sample is facing positive 90
        in the azimuthal direction when positively tilted.

    Returns
    -------
    take_off_angle: float.
        In degrees.

    Examples
    --------
    >>> hs.eds.take_off_angle(alpha_tilt=10., beta_tilt=0.
    >>>                          azimuth_angle=45., elevation_angle=22.)
    28.865971201155283
    """

    if tilt_stage is None:
        raise ValueError(
            "Unable to calculate take-off angle. The metadata property "
            "`Stage.tilt_alpha` is not set."
        )

    if azimuth_angle is None:
        raise ValueError(
            "Unable to calculate take-off angle. The metadata property "
            "`Detector.EDS.azimuth_angle` is not set."
        )

    if elevation_angle is None:
        raise ValueError(
            "Unable to calculate take-off angle. The metadata property "
            "`Detector.EDS.elevation_angle` is not set."
        )

    alpha = math.radians(tilt_stage)
    beta = -math.radians(beta_tilt)
    phi = math.radians(azimuth_angle)
    theta = -math.radians(elevation_angle)

    return 90 - math.degrees(
        np.arccos(
            math.sin(alpha) * math.cos(beta) * math.cos(phi) * math.cos(theta)
            - math.sin(beta) * math.sin(phi) * math.cos(theta)
            - math.cos(alpha) * math.cos(beta) * math.sin(theta)
        )
    )


def xray_lines_model(
    elements,
    beam_energy=200,
    weight_percents=None,
    energy_resolution_MnKa=130,
    energy_axis=None,
):
    """
    Generate a model of X-ray lines using a Gaussian distribution for each
    peak.

    The area under a main peak (alpha) is equal to 1 and weighted by the
    composition.

    Parameters
    ----------
    elements : list of strings
        A list of chemical element symbols.
    beam_energy: float
        The energy of the beam in keV.
    weight_percents: list of float
        The composition in weight percent.
    energy_resolution_MnKa: float
        The energy resolution of the detector in eV
    energy_axis: dic
        The dictionary for the energy axis. It must contains 'size' and the
        units must be 'eV' of 'keV'.

    Example
    -------
    >>> s = xray_lines_model(['Cu', 'Fe'], beam_energy=30)
    >>> s.plot()
    """
    from exspy.signals.eds_tem import EDSTEMSpectrum
    from hyperspy import components1d

    if energy_axis is None:
        energy_axis = {
            "name": "E",
            "scale": 0.01,
            "units": "keV",
            "offset": -0.1,
            "size": 1024,
        }
    s = EDSTEMSpectrum(np.zeros(energy_axis["size"]), axes=[energy_axis])
    s.set_microscope_parameters(
        beam_energy=beam_energy, energy_resolution_MnKa=energy_resolution_MnKa
    )
    s.add_elements(elements)
    counts_rate = 1.0
    live_time = 1.0
    if weight_percents is None:
        weight_percents = [100.0 / len(elements)] * len(elements)
    m = s.create_model()
    if len(elements) == len(weight_percents):
        for element, weight_percent in zip(elements, weight_percents):
            for line, properties in elements_db[element]["Atomic_properties"][
                "Xray_lines"
            ].items():
                line_energy = properties["energy (keV)"]
                ratio_line = properties["weight"]
                if s._get_xray_lines_in_spectral_range([element + "_" + line])[1] == []:
                    g = components1d.Gaussian()
                    g.centre.value = line_energy
                    g.sigma.value = (
                        get_FWHM_at_Energy(energy_resolution_MnKa, line_energy)
                        / sigma2fwhm
                    )
                    g.A.value = (
                        live_time * counts_rate * weight_percent / 100 * ratio_line
                    )
                    m.append(g)
    else:
        raise ValueError(
            "The number of elements specified is not the same "
            "as the number of weight_percents"
        )

    s.data = m.as_signal().data
    return s


def quantification_cliff_lorimer(
    intensities, kfactors, absorption_correction=None, mask=None
):
    """
    Quantification using Cliff-Lorimer

    Parameters
    ----------
    intensities: numpy.array
        the intensities for each X-ray lines. The first axis should be the
        elements axis.
    kfactors: list of float
        The list of kfactor in same order as intensities eg. kfactors =
        [1, 1.47, 1.72] for ['Al_Ka','Cr_Ka', 'Ni_Ka']
    %s
    mask: array of bool, signal of bool or None
        The mask with the dimension of intensities[0]. If a pixel is True,
        the composition is set to zero.

    Return
    ------
    numpy.array containing the weight fraction with the same
    shape as intensities.
    """
    # Value used as an threshold to prevent using zeros as denominator
    min_intensity = 0.1
    dim = intensities.shape
    dim2 = reduce(lambda x, y: x * y, dim[1:])
    intens = intensities.reshape(dim[0], dim2).astype(float)

    if absorption_correction is None:
        # default to ones
        absorption_correction = np.ones_like(intens, dtype=float)
    else:
        absorption_correction = absorption_correction.reshape(dim[0], dim2)

    for i in range(dim2):
        index = np.where(intens[:, i] > min_intensity)[0]
        if len(index) > 1:
            ref_index, ref_index2 = index[:2]
            intens[:, i] = _quantification_cliff_lorimer(
                intens[:, i],
                kfactors,
                absorption_correction[:, i],
                ref_index,
                ref_index2,
            )
        else:
            intens[:, i] = np.zeros_like(intens[:, i])
            if len(index) == 1:
                intens[index[0], i] = 1.0

    intens = intens.reshape(dim)
    if mask is not None:
        from hyperspy.signals import BaseSignal

        if isinstance(mask, BaseSignal):
            mask = mask.data
        for i in range(dim[0]):
            intens[i][(mask == True)] = 0

    return intens


quantification_cliff_lorimer.__doc__ %= _ABSORPTION_CORRECTION_DOCSTRING


def _quantification_cliff_lorimer(
    intensities, kfactors, absorption_correction, ref_index=0, ref_index2=1
):
    """
    Quantification using Cliff-Lorimer

    Parameters
    ----------
    intensities: numpy.array
        the intensities for each X-ray lines. The first axis should be the
        elements axis.
    absorption_correction: numpy.ndarray
        value between 0 and 1 in order to correct the intensities based on
        estimated absorption.
    kfactors: list of float
        The list of kfactor in same order as  intensities eg. kfactors =
        [1, 1.47, 1.72] for ['Al_Ka','Cr_Ka', 'Ni_Ka']
    ref_index, ref_index2: int
        index of the elements that will be in the denominator. Should be non
        zeros if possible.

    Return
    ------
    numpy.array containing the weight fraction with the same
    shape as intensities.
    """
    if len(intensities) != len(kfactors):
        raise ValueError(
            "The number of kfactors must match the size of the "
            "first axis of intensities."
        )

    ab = np.zeros_like(intensities, dtype="float")
    composition = np.ones_like(intensities, dtype="float")
    # ab = Ia/Ib / kab
    other_index = list(range(len(kfactors)))
    other_index.pop(ref_index)
    for i in other_index:
        ab[i] = (
            (intensities[ref_index] * absorption_correction[ref_index])
            / (intensities[i] * absorption_correction[i])
            * (kfactors[ref_index] / kfactors[i])
        )
    # Ca = ab /(1 + ab + ab/ac + ab/ad + ...)
    for i in other_index:
        if i == ref_index2:
            composition[ref_index] += ab[ref_index2]
        else:
            composition[ref_index] += ab[ref_index2] / ab[i]
    composition[ref_index] = ab[ref_index2] / composition[ref_index]
    # Cb = Ca / ab
    for i in other_index:
        composition[i] = composition[ref_index] / ab[i]
    return composition


def quantification_zeta_factor(intensities, zfactors, dose, absorption_correction=None):
    """
    Quantification using the zeta-factor method

    Parameters
    ----------
    intensities: numpy.array
        The intensities for each X-ray line. The first axis should be the
        elements axis.
    zfactors: list of float
        The list of zeta-factors in the same order as intensities
        e.g. zfactors = [628.10, 539.89] for ['As_Ka', 'Ga_Ka'].
    dose: float
        The total electron dose given by i*t*N, i the current,
        t the acquisition time and
        N the number of electrons per unit electric charge (1/e).
    %s

    Returns
    ------
    A numpy.array containing the weight fraction with the same
    shape as intensities and mass thickness in kg/m^2.
    """
    if absorption_correction is None:
        # default to ones
        absorption_correction = np.ones_like(intensities, dtype="float")

    sumzi = np.zeros_like(intensities[0], dtype="float")
    composition = np.zeros_like(intensities, dtype="float")
    for intensity, zfactor, acf in zip(intensities, zfactors, absorption_correction):
        sumzi = sumzi + (intensity * zfactor * acf)
    for i, (intensity, zfactor, acf) in enumerate(
        zip(intensities, zfactors, absorption_correction)
    ):
        composition[i] = intensity * zfactor * acf / sumzi
    mass_thickness = sumzi / dose
    return composition, mass_thickness


quantification_zeta_factor.__doc__ %= _ABSORPTION_CORRECTION_DOCSTRING


def get_abs_corr_zeta(weight_percent, mass_thickness, take_off_angle):
    """
    Calculate absorption correction terms.

    Parameters
    ----------
    weight_percent: list of signal
        Composition in weight percent.
    mass_thickness: signal
        Density-thickness map in kg/m^2
    take_off_angle: float
        X-ray take-off angle in degrees.
    """
    from exspy.misc import material

    toa_rad = np.radians(take_off_angle)
    csc_toa = 1.0 / np.sin(toa_rad)
    # convert from cm^2/g to m^2/kg
    mac = (
        stack(
            material.mass_absorption_mixture(weight_percent=weight_percent),
            show_progressbar=False,
        )
        * 0.1
    )
    expo = mac.data * mass_thickness.data * csc_toa
    acf = expo / (1.0 - np.exp(-(expo)))
    return acf


def quantification_cross_section(
    intensities, cross_sections, dose, absorption_correction=None
):
    """
    Quantification using EDX cross sections
    Calculate the atomic compostion and the number of atoms per pixel
    from the raw X-ray intensity

    Parameters
    ----------
    intensity : numpy.ndarray
        The integrated intensity for each X-ray line, where the first axis
        is the element axis.
    cross_sections : list of floats
        List of X-ray scattering cross-sections in the same order as the
        intensities.
    dose: float
        the dose per unit area given by i*t*N/A, i the current,
        t the acquisition time, and
        N the number of electron by unit electric charge.
    %s

    Returns
    -------
    numpy.array containing the atomic fraction of each element, with
    the same shape as the intensity input.
    numpy.array of the number of atoms counts for each element, with the same
    shape as the intensity input.
    """

    if absorption_correction is None:
        # default to ones
        absorption_correction = np.ones_like(intensities, dtype=float)

    shp = len(intensities.shape) - 1
    slices = (slice(None),) + (None,) * shp
    x_sections = np.array(cross_sections, dtype=float)[slices]
    number_of_atoms = intensities / (x_sections * dose * 1e-10) * absorption_correction
    total_atoms = np.cumsum(number_of_atoms, axis=0)[-1]
    composition = number_of_atoms / total_atoms

    return composition, number_of_atoms


quantification_cross_section.__doc__ %= _ABSORPTION_CORRECTION_DOCSTRING


def get_abs_corr_cross_section(
    composition, number_of_atoms, take_off_angle, probe_area
):
    """
    Calculate absorption correction terms.

    Parameters
    ----------
    number_of_atoms: list of signal
        Stack of maps with number of atoms per pixel.
    take_off_angle: float
        X-ray take-off angle in degrees.
    """
    from exspy.misc import material

    toa_rad = np.radians(take_off_angle)
    Av = constants.Avogadro
    elements = [intensity.metadata.Sample.elements[0] for intensity in number_of_atoms]
    atomic_weights = np.array(
        [
            elements_db[element]["General_properties"]["atomic_weight"]
            for element in elements
        ]
    )

    number_of_atoms = stack(number_of_atoms, show_progressbar=False).data

    # calculate the total_mass in kg/m^2, or mass thickness.
    total_mass = np.zeros_like(number_of_atoms[0], dtype="float")
    for i, (weight) in enumerate(atomic_weights):
        total_mass += number_of_atoms[i] * weight / Av / 1e3 / probe_area / 1e-18
    # determine mass absorption coefficients and convert from cm^2/g to m^2/kg.
    to_stack = material.mass_absorption_mixture(
        weight_percent=material.atomic_to_weight(composition)
    )
    mac = stack(to_stack, show_progressbar=False) * 0.1
    acf = np.zeros_like(number_of_atoms)
    csc_toa = 1 / math.sin(toa_rad)
    # determine an absorption coeficient per element per pixel.
    for i, (weight) in enumerate(atomic_weights):
        expo = mac.data[i] * total_mass * csc_toa
        acf[i] = expo / (1 - np.exp(-expo))
    return acf


def edx_cross_section_to_zeta(cross_sections, elements):
    """Convert a list of cross_sections in barns (b) to zeta-factors (kg/m^2).

    Parameters
    ----------
    cross_section: list of float
        A list of cross sections in barns.
    elements: list of str
        A list of element chemical symbols in the same order as the
        cross sections e.g. ['Al','Zn']

    Returns
    -------
    zeta_factors : list of float
        zeta_factors with units kg/m^2.

    """
    if len(elements) != len(cross_sections):
        raise ValueError(
            "The number of elements must match the number of cross sections."
        )
    zeta_factors = []
    for i, element in enumerate(elements):
        atomic_weight = elements_db[element]["General_properties"]["atomic_weight"]
        zeta = atomic_weight / (cross_sections[i] * constants.Avogadro * 1e-25)
        zeta_factors.append(zeta)
    return zeta_factors


def zeta_to_edx_cross_section(zfactors, elements):
    """Convert a list of zeta-factors (kg/m^2) to cross_sections in barns (b).

    Parameters
    ----------
    zfactors: list of float
        A list of zeta-factors.
    elements: list of str
        A list of element chemical symbols in the same order as the
        cross sections e.g. ['Al','Zn']

    Returns
    -------
    cross_sections : list of float
        cross_sections with units in barns.

    """
    if len(elements) != len(zfactors):
        raise ValueError(
            "The number of elements must match the number of cross sections."
        )
    cross_sections = []
    for i, element in enumerate(elements):
        atomic_weight = elements_db[element]["General_properties"]["atomic_weight"]
        xsec = atomic_weight / (zfactors[i] * constants.Avogadro * 1e-25)
        cross_sections.append(xsec)
    return cross_sections



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The HyperSpy developers
#
# This file is part of HyperSpy.
#
# HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

#
# You should have received a copy of the GNU General Public License
# along with HyperSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import itertools

import numpy as np
import pytest

from hyperspy.exceptions import VisibleDeprecationWarning
from hyperspy.signals import Signal1D

from exspy.components import PESVoigt


TRUE_FALSE_2_TUPLE = [p for p in itertools.product((True, False), repeat=2)]


def test_function():
    g = PESVoigt()
    g.area.value = 5
    g.FWHM.value = 0.5
    g.gamma.value = 0.2
    g.centre.value = 1
    np.testing.assert_allclose(g.function(0), 0.35380168)
    np.testing.assert_allclose(g.function(1), 5.06863535)
    assert g._position is g.centre


def test_function_resolution():
    g = PESVoigt()
    g.area.value = 5
    g.FWHM.value = 0.5
    g.gamma.value = 0.2
    g.centre.value = 1
    g.resolution.value = 0.7
    np.testing.assert_allclose(g.function(0), 0.5335443)
    np.testing.assert_allclose(g.function(1), 3.70472923)
    assert g._position is g.centre


def test_function_spinorbit():
    g = PESVoigt()
    g.area.value = 5
    g.FWHM.value = 0.5
    g.gamma.value = 0.2
    g.centre.value = 1
    g.spin_orbit_splitting = True
    spin_orbit_branching_ratio = 0.4
    spin_orbit_splitting_energy = 0.72
    np.testing.assert_allclose(g.function(0), 1.553312)
    np.testing.assert_allclose(g.function(1), 5.612734)
    assert g._position is g.centre


def test_function_shirleybackground():
    g = PESVoigt()
    g.area.value = 5
    g.FWHM.value = 0.5
    g.gamma.value = 0.2
    g.centre.value = 1
    g.shirley_background.value = 1.5
    g.shirley_background.active = True
    np.testing.assert_allclose(g.function(0), 0.35380168)
    np.testing.assert_allclose(g.function(1), 5.06863535)
    assert g._position is g.centre


@pytest.mark.parametrize(("lazy"), (True, False))
@pytest.mark.parametrize(("uniform"), (True, False))
@pytest.mark.parametrize(("mapnone"), (True, False))
@pytest.mark.parametrize(("only_current", "binned"), TRUE_FALSE_2_TUPLE)
def test_estimate_parameters_binned(only_current, binned, lazy, uniform, mapnone):
    s = Signal1D(np.empty((200,)))
    s.axes_manager.signal_axes[0].is_binned = binned
    axis = s.axes_manager.signal_axes[0]
    axis.scale = 0.05
    axis.offset = -5
    g1 = PESVoigt()
    g1.centre.value = 1
    g1.area.value = 5.0
    g1.gamma.value = 0.001
    g1.FWHM.value = 0.5
    s.data = g1.function(axis.axis)
    if not uniform:
        axis.convert_to_non_uniform_axis()
    if lazy:
        s = s.as_lazy()
    g2 = PESVoigt()
    if binned and uniform:
        factor = axis.scale
    elif binned:
        factor = np.gradient(axis.axis)
    else:
        factor = 1
    if mapnone:
        g2.area.map = None
    assert g2.estimate_parameters(
        s, axis.low_value, axis.high_value, only_current=only_current
    )
    assert g2._axes_manager[-1].is_binned == binned
    np.testing.assert_allclose(g2.FWHM.value, 1, 0.5)
    np.testing.assert_allclose(g1.area.value, g2.area.value * factor, 0.04)
    np.testing.assert_allclose(g2.centre.value, 1, 1e-3)



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.


import math


def effective_angle(E0, E, alpha, beta):
    """Calculates the effective collection angle

    Parameters
    ----------
    E0 : float
        electron beam energy in keV
    E : float
        energy loss in eV
    alpha : float
        convergence angle in mrad
    beta : float
        collection angle in mrad

    Returns
    -------
    float : effective collection angle in mrad

    Notes
    -----
    Code translated to Python from Egerton (second edition) page 420

    """
    E0 *= 1e3  # keV to eV
    if alpha == 0:
        return beta
    E0 *= 10.0**-3  # In KeV
    E = float(E)
    alpha = float(alpha)
    beta = float(beta)
    TGT = E0 * (1.0 + E0 / 1022.0) / (1.0 + E0 / 511.0)
    thetaE = E / TGT
    A2 = alpha * alpha * 1e-6
    B2 = beta * beta * 1e-6
    T2 = thetaE * thetaE * 1e-6
    eta1 = math.sqrt((A2 + B2 + T2) ** 2 - 4.0 * A2 * B2) - A2 - B2 - T2
    eta2 = (
        2.0
        * B2
        * math.log(
            0.5 / T2 * (math.sqrt((A2 + T2 - B2) ** 2 + 4.0 * B2 * T2) + A2 + T2 - B2)
        )
    )
    eta3 = (
        2.0
        * A2
        * math.log(
            0.5 / T2 * (math.sqrt((B2 + T2 - A2) ** 2 + 4.0 * A2 * T2) + B2 + T2 - A2)
        )
    )
    #    ETA=(eta1+eta2+eta3)/A2/math.log(4./T2)
    F1 = (eta1 + eta2 + eta3) / 2 / A2 / math.log(1.0 + B2 / T2)
    F2 = F1
    if (alpha / beta) > 1:
        F2 = F1 * A2 / B2
    BSTAR = thetaE * math.sqrt(math.exp(F2 * math.log(1.0 + B2 / T2)) - 1.0)
    return BSTAR  # In mrad



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import pytest

import numpy as np

from hyperspy import utils
from hyperspy.components1d import Gaussian
from hyperspy.decorators import lazifyTestClass

import exspy
from exspy._defaults_parser import preferences
import exspy.misc.eds.utils as eds_utils
from exspy.signals import EDSSEMSpectrum


@lazifyTestClass
class Test_metadata:
    def setup_method(self, method):
        # Create an empty spectrum
        s = EDSSEMSpectrum(np.ones((4, 2, 1024)))
        s.axes_manager.signal_axes[0].scale = 1e-3
        s.axes_manager.signal_axes[0].units = "keV"
        s.axes_manager.signal_axes[0].name = "Energy"
        s.metadata.Acquisition_instrument.SEM.Detector.EDS.live_time = 3.1
        s.metadata.Acquisition_instrument.SEM.beam_energy = 15.0
        s.metadata.Acquisition_instrument.SEM.Stage.tilt_alpha = -38
        s.metadata.Acquisition_instrument.SEM.Detector.EDS.azimuth_angle = 63
        s.metadata.Acquisition_instrument.SEM.Detector.EDS.elevation_angle = 35
        self.signal = s

    def test_sum_live_time(self):
        s = self.signal
        old_metadata = s.metadata.deepcopy()
        sSum = s.sum(0)
        assert (
            sSum.metadata.Acquisition_instrument.SEM.Detector.EDS.live_time
            == s.metadata.Acquisition_instrument.SEM.Detector.EDS.live_time * 2
        )
        # Check that metadata is unchanged
        print(old_metadata, s.metadata)  # Capture for comparison on error
        assert (
            old_metadata.as_dictionary() == s.metadata.as_dictionary()
        ), "Source metadata changed"

    def test_sum_live_time2(self):
        s = self.signal
        old_metadata = s.metadata.deepcopy()
        sSum = s.sum((0, 1))
        assert (
            sSum.metadata.Acquisition_instrument.SEM.Detector.EDS.live_time
            == s.metadata.Acquisition_instrument.SEM.Detector.EDS.live_time * 2 * 4
        )
        # Check that metadata is unchanged
        print(old_metadata, s.metadata)  # Capture for comparison on error
        assert (
            old_metadata.as_dictionary() == s.metadata.as_dictionary()
        ), "Source metadata changed"

    def test_sum_live_time_out_arg(self):
        s = self.signal
        sSum = s.sum(0)
        sSum.metadata.Acquisition_instrument.SEM.Detector.EDS.live_time = 4.2
        s_resum = s.sum(0)
        r = s.sum(0, out=sSum)
        assert r is None
        assert (
            s_resum.metadata.Acquisition_instrument.SEM.Detector.EDS.live_time
            == s.metadata.Acquisition_instrument.SEM.Detector.EDS.live_time * 2
        )
        np.testing.assert_allclose(s_resum.data, sSum.data)

    def test_rebin_live_time(self):
        s = self.signal
        old_metadata = s.metadata.deepcopy()
        s = s.rebin(scale=[2, 2, 1])
        assert (
            s.metadata.Acquisition_instrument.SEM.Detector.EDS.live_time == 3.1 * 2 * 2
        )
        # Check that metadata is unchanged
        print(old_metadata, self.signal.metadata)  # Captured on error
        assert (
            old_metadata.as_dictionary() == self.signal.metadata.as_dictionary()
        ), "Source metadata changed"

    def test_add_elements(self):
        s = self.signal
        s.add_elements(["Al", "Ni"])
        assert s.metadata.Sample.elements == ["Al", "Ni"]
        s.add_elements(["Al", "Ni"])
        assert s.metadata.Sample.elements == ["Al", "Ni"]
        s.add_elements(
            [
                "Fe",
            ]
        )
        assert s.metadata.Sample.elements == ["Al", "Fe", "Ni"]
        s.set_elements(["Al", "Ni"])
        assert s.metadata.Sample.elements == ["Al", "Ni"]

    def test_add_lines(self):
        s = self.signal
        s.add_lines(lines=())
        assert s.metadata.Sample.xray_lines == []
        s.add_lines(("Fe_Ln",))
        assert s.metadata.Sample.xray_lines == ["Fe_Ln"]
        s.add_lines(("Fe_Ln",))
        assert s.metadata.Sample.xray_lines == ["Fe_Ln"]
        s.add_elements(
            [
                "Ti",
            ]
        )
        s.add_lines(())
        assert s.metadata.Sample.xray_lines == ["Fe_Ln", "Ti_La"]
        s.set_lines((), only_one=False, only_lines=False)
        assert s.metadata.Sample.xray_lines == [
            "Fe_La",
            "Fe_Lb3",
            "Fe_Ll",
            "Fe_Ln",
            "Ti_La",
            "Ti_Lb3",
            "Ti_Ll",
            "Ti_Ln",
        ]
        s.metadata.Acquisition_instrument.SEM.beam_energy = 0.4
        s.set_lines((), only_one=False, only_lines=False)
        assert s.metadata.Sample.xray_lines == ["Ti_Ll"]

    def test_add_lines_warning(self):
        s = self.signal
        with pytest.warns(UserWarning):
            s.add_lines(("Fe_Ka",))

    def test_add_lines_auto(self):
        s = self.signal
        s.axes_manager.signal_axes[0].scale = 1e-2
        s.set_elements(["Ti", "Al"])
        s.set_lines(["Al_Ka"])
        assert s.metadata.Sample.xray_lines == ["Al_Ka", "Ti_Ka"]

        del s.metadata.Sample.xray_lines
        s.set_elements(["Al", "Ni"])
        s.add_lines()
        assert s.metadata.Sample.xray_lines == ["Al_Ka", "Ni_Ka"]
        s.metadata.Acquisition_instrument.SEM.beam_energy = 10.0
        s.set_lines([])
        assert s.metadata.Sample.xray_lines == ["Al_Ka", "Ni_La"]
        s.metadata.Acquisition_instrument.SEM.beam_energy = 200
        s.set_elements(["Au", "Ni"])
        s.set_lines([])
        assert s.metadata.Sample.xray_lines == ["Au_La", "Ni_Ka"]

    def test_default_param(self):
        s = self.signal
        mp = s.metadata
        assert (
            mp.Acquisition_instrument.SEM.Detector.EDS.energy_resolution_MnKa
            == preferences.EDS.eds_mn_ka
        )

    def test_SEM_to_TEM(self):
        s = self.signal.inav[0, 0]
        signal_type = "EDS_TEM"
        mp = s.metadata
        mp.Acquisition_instrument.SEM.Detector.EDS.energy_resolution_MnKa = 125.3
        sTEM = s.deepcopy()
        sTEM.set_signal_type(signal_type)
        mpTEM = sTEM.metadata
        results = [
            mp.Acquisition_instrument.SEM.Detector.EDS.energy_resolution_MnKa,
            signal_type,
        ]
        resultsTEM = [
            (mpTEM.Acquisition_instrument.TEM.Detector.EDS.energy_resolution_MnKa),
            mpTEM.Signal.signal_type,
        ]
        assert results == resultsTEM

    def test_get_calibration_from(self):
        s = self.signal
        scalib = EDSSEMSpectrum(np.ones(1024))
        energy_axis = scalib.axes_manager.signal_axes[0]
        energy_axis.scale = 0.01
        energy_axis.offset = -0.10
        s.get_calibration_from(scalib)
        assert s.axes_manager.signal_axes[0].scale == energy_axis.scale

    def test_take_off_angle(self):
        s = self.signal
        np.testing.assert_allclose(s.get_take_off_angle(), 12.886929785732487)

    def test_take_off_angle_beta(self):
        s = self.signal
        s.metadata.Acquisition_instrument.SEM.Stage.tilt_beta = 15.5
        np.testing.assert_allclose(s.get_take_off_angle(), 24.202671071140102)

    def test_take_off_angle_alpha(self):
        s = self.signal
        s.metadata.Acquisition_instrument.SEM.Stage.tilt_alpha = None
        with pytest.raises(ValueError, match="alpha"):
            s.get_take_off_angle()

    def test_take_off_angle_azimuth(self):
        s = self.signal
        s.metadata.Acquisition_instrument.SEM.Detector.EDS.azimuth_angle = None
        with pytest.raises(ValueError, match="azimuth"):
            s.get_take_off_angle()

    def test_take_off_angle_elevation(self):
        s = self.signal
        s.metadata.Acquisition_instrument.SEM.Detector.EDS.elevation_angle = None
        with pytest.raises(ValueError, match="elevation"):
            s.get_take_off_angle()


@lazifyTestClass
class Test_get_lines_intensity:
    def setup_method(self, method):
        # Create an empty spectrum
        s = EDSSEMSpectrum(np.zeros((2, 2, 3, 100)))
        energy_axis = s.axes_manager.signal_axes[0]
        energy_axis.scale = 0.04
        energy_axis.units = "keV"
        energy_axis.name = "Energy"
        g = Gaussian()
        g.sigma.value = 0.05
        g.centre.value = 1.487
        s.data[:] = g.function(energy_axis.axis)
        s.metadata.Acquisition_instrument.SEM.Detector.EDS.live_time = 3.1
        s.metadata.Acquisition_instrument.SEM.beam_energy = 15.0
        self.signal = s

    @pytest.mark.parametrize("bad_iter", ["Al_Kb", {"A": "Al_Kb", "B": "Ca_Ka"}])
    def test_bad_iter(self, bad_iter):
        # get_lines_intensity() should raise TypeError when
        # xray_lines is a string or a dictionary
        s = self.signal

        with pytest.raises(TypeError):
            s.get_lines_intensity(xray_lines=bad_iter, plot_result=False)

    @pytest.mark.parametrize(
        "good_iter", [("Al_Kb", "Ca_Ka"), ["Al_Kb", "Ca_Ka"], set(["Al_Kb", "Ca_Ka"])]
    )
    def test_good_iter(self, good_iter):
        s = self.signal

        # get_lines_intensity() should succeed and return a list
        # when xray_lines is an iterable (other than a str or dict)
        assert isinstance(
            s.get_lines_intensity(xray_lines=good_iter, plot_result=False), list
        )

    def test(self):
        s = self.signal

        sAl = s.get_lines_intensity(
            ["Al_Ka"], plot_result=False, integration_windows=5
        )[0]
        assert sAl.axes_manager.signal_dimension == 0
        np.testing.assert_allclose(
            sAl.metadata.Acquisition_instrument.SEM.Detector.EDS.live_time,
            s.metadata.Acquisition_instrument.SEM.Detector.EDS.live_time,
        )
        np.testing.assert_allclose(24.99516, sAl.data[0, 0, 0], atol=1e-3)
        sAl = s.inav[0].get_lines_intensity(
            ["Al_Ka"], plot_result=False, integration_windows=5
        )[0]
        np.testing.assert_allclose(24.99516, sAl.data[0, 0], atol=1e-3)
        sAl = s.inav[0, 0].get_lines_intensity(
            ["Al_Ka"], plot_result=False, integration_windows=5
        )[0]
        np.testing.assert_allclose(24.99516, sAl.data[0], atol=1e-3)
        sAl = s.inav[0, 0, 0].get_lines_intensity(
            ["Al_Ka"], plot_result=False, integration_windows=5
        )[0]
        np.testing.assert_allclose(24.99516, sAl.data, atol=1e-3)
        s.axes_manager[-1].offset = 1.0
        with pytest.warns(UserWarning, match="C_Ka is not in the data energy range."):
            sC = s.get_lines_intensity(["C_Ka"], plot_result=False)
        assert len(sC) == 0
        assert sAl.metadata.Sample.elements == ["Al"]
        assert sAl.metadata.Sample.xray_lines == ["Al_Ka"]

    def test_eV(self):
        s = self.signal
        energy_axis = s.axes_manager.signal_axes[0]
        energy_axis.scale = 40
        energy_axis.units = "eV"

        sAl = s.get_lines_intensity(
            ["Al_Ka"], plot_result=False, integration_windows=5
        )[0]
        np.testing.assert_allclose(24.99516, sAl.data[0, 0, 0], atol=1e-3)

    def test_plot_result_single_spectrum(self):
        s = self.signal.inav[0, 0, 0]
        intensities = s.get_lines_intensity(["Al_Ka"], plot_result=True)
        assert intensities[0].data.size == 1

    def test_background_substraction(self):
        s = self.signal
        intens = s.get_lines_intensity(["Al_Ka"], plot_result=False)[0].data
        s = s + 1.0
        np.testing.assert_allclose(
            s.estimate_background_windows(xray_lines=["Al_Ka"])[0, 0],
            1.25666201,
            atol=1e-3,
        )
        np.testing.assert_allclose(
            s.get_lines_intensity(
                ["Al_Ka"],
                background_windows=s.estimate_background_windows(
                    [4, 4], xray_lines=["Al_Ka"]
                ),
                plot_result=False,
            )[0].data,
            intens,
            atol=1e-3,
        )

    def test_estimate_integration_windows(self):
        s = self.signal
        np.testing.assert_allclose(
            s.estimate_integration_windows(3.0, ["Al_Ka"]), [[1.371, 1.601]], atol=1e-2
        )

    def test_with_signals_examples(self):
        s = exspy.data.EDS_SEM_TM002()
        np.testing.assert_allclose(
            utils.stack(s.get_lines_intensity()).data.squeeze(),
            np.array([84163, 89063, 96117, 96700, 99075]),
        )


@lazifyTestClass
class Test_tools_bulk:
    def setup_method(self, method):
        s = EDSSEMSpectrum(np.ones(1024))
        s.metadata.Acquisition_instrument.SEM.beam_energy = 5.0
        energy_axis = s.axes_manager.signal_axes[0]
        energy_axis.scale = 0.01
        energy_axis.units = "keV"
        s.set_elements(["Al", "Zn"])
        s.add_lines()
        self.signal = s

    def test_electron_range(self):
        s = self.signal
        mp = s.metadata
        elec_range = eds_utils.electron_range(
            mp.Sample.elements[0],
            mp.Acquisition_instrument.SEM.beam_energy,
            density="auto",
            tilt=mp.Acquisition_instrument.SEM.Stage.tilt_alpha,
        )
        np.testing.assert_allclose(elec_range, 0.41350651162374225)

    def test_xray_range(self):
        s = self.signal
        mp = s.metadata
        xr_range = eds_utils.xray_range(
            mp.Sample.xray_lines[0],
            mp.Acquisition_instrument.SEM.beam_energy,
            density=4.37499648818,
        )
        np.testing.assert_allclose(xr_range, 0.1900368800933955)


@lazifyTestClass
class Test_energy_units:
    def setup_method(self, method):
        s = EDSSEMSpectrum(np.ones(1024))
        s.metadata.Acquisition_instrument.SEM.beam_energy = 5.0
        s.axes_manager.signal_axes[0].units = "keV"
        s.set_microscope_parameters(energy_resolution_MnKa=130)
        self.signal = s

    def test_beam_energy(self):
        s = self.signal
        assert s._get_beam_energy() == 5.0
        s.axes_manager.signal_axes[0].units = "eV"
        assert s._get_beam_energy() == 5000.0
        s.axes_manager.signal_axes[0].units = "keV"

    def test_line_energy(self):
        s = self.signal
        assert s._get_line_energy("Al_Ka") == 1.4865
        s.axes_manager.signal_axes[0].units = "eV"
        assert s._get_line_energy("Al_Ka") == 1486.5
        s.axes_manager.signal_axes[0].units = "keV"

        np.testing.assert_allclose(
            s._get_line_energy("Al_Ka", FWHM_MnKa="auto"), (1.4865, 0.07661266213883969)
        )
        np.testing.assert_allclose(
            s._get_line_energy("Al_Ka", FWHM_MnKa=128), (1.4865, 0.073167615787314)
        )



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.


import numpy as np
import pytest
from pathlib import Path

import hyperspy.api as hs
from hyperspy.decorators import lazifyTestClass

import exspy


MYPATH = Path(__file__).resolve().parent


@lazifyTestClass
class Test_Estimate_Elastic_Scattering_Threshold:
    def setup_method(self, method):
        # Create an empty spectrum
        s = exspy.signals.EELSSpectrum(np.zeros((3, 2, 1024)))
        energy_axis = s.axes_manager.signal_axes[0]
        energy_axis.scale = 0.02
        energy_axis.offset = -5

        gauss = hs.model.components1D.Gaussian()
        gauss.centre.value = 0
        gauss.A.value = 5000
        gauss.sigma.value = 0.5
        gauss2 = hs.model.components1D.Gaussian()
        gauss2.sigma.value = 0.5
        # Inflexion point 1.5
        gauss2.A.value = 5000
        gauss2.centre.value = 5
        s.data[:] = gauss.function(energy_axis.axis) + gauss2.function(energy_axis.axis)
        self.signal = s

    def test_min_in_window_with_smoothing(self):
        s = self.signal
        thr = s.estimate_elastic_scattering_threshold(
            window=5,
            window_length=5,
            tol=0.00001,
        )
        np.testing.assert_allclose(thr.data, 2.5, atol=10e-3)
        assert thr.metadata.Signal.signal_type == ""
        assert thr.axes_manager.signal_dimension == 0

    def test_min_in_window_without_smoothing_single_spectrum(self):
        s = self.signal.inav[0, 0]
        thr = s.estimate_elastic_scattering_threshold(
            window=5,
            window_length=0,
            tol=0.001,
        )
        np.testing.assert_allclose(thr.data, 2.49, atol=10e-3)

    def test_min_in_window_without_smoothing(self):
        s = self.signal
        thr = s.estimate_elastic_scattering_threshold(
            window=5,
            window_length=0,
            tol=0.001,
        )
        np.testing.assert_allclose(thr.data, 2.49, atol=10e-3)

    def test_min_not_in_window(self):
        # If I use a much lower window, this is the value that has to be
        # returned as threshold.
        s = self.signal
        data = s.estimate_elastic_scattering_threshold(
            window=1.5,
            tol=0.001,
        ).data
        assert np.all(np.isnan(data))

    def test_estimate_elastic_scattering_intensity(self):
        s = self.signal
        threshold = s.estimate_elastic_scattering_threshold(window=4.0)
        # Threshold is nd signal
        t = s.estimate_elastic_scattering_intensity(threshold=threshold)
        assert t.metadata.Signal.signal_type == ""
        assert t.axes_manager.signal_dimension == 0
        np.testing.assert_array_almost_equal(t.data, 249999.985133)
        # Threshold is signal, 1 spectrum
        s0 = s.inav[0]
        t0 = s0.estimate_elastic_scattering_threshold(window=4.0)
        t = s0.estimate_elastic_scattering_intensity(threshold=t0)
        np.testing.assert_array_almost_equal(t.data, 249999.985133)
        # Threshold is value
        t = s.estimate_elastic_scattering_intensity(threshold=2.5)
        np.testing.assert_array_almost_equal(t.data, 249999.985133)


@lazifyTestClass
class TestEstimateZLPCentre:
    def setup_method(self, method):
        s = exspy.signals.EELSSpectrum(np.diag(np.arange(1, 11)))
        s.axes_manager[-1].scale = 0.1
        s.axes_manager[-1].offset = 100
        self.signal = s

    def test_estimate_zero_loss_peak_centre(self):
        s = self.signal
        zlpc = s.estimate_zero_loss_peak_centre()
        np.testing.assert_allclose(zlpc.data, np.arange(100, 101, 0.1))
        assert zlpc.metadata.Signal.signal_type == ""
        assert zlpc.axes_manager.signal_dimension == 0


@lazifyTestClass
class TestAlignZLP:
    def setup_method(self, method):
        s = exspy.signals.EELSSpectrum(np.zeros((10, 100)))
        self.scale = 0.1
        self.offset = -2
        eaxis = s.axes_manager.signal_axes[0]
        eaxis.scale = self.scale
        eaxis.offset = self.offset
        self.izlp = eaxis.value2index(0)
        self.bg = 2
        self.ishifts = np.array([0, 4, 2, -2, 5, -2, -5, -9, -9, -8])
        self.new_offset = self.offset - self.ishifts.min() * self.scale
        s.data[np.arange(10), self.ishifts + self.izlp] = 10
        s.data += self.bg
        s.axes_manager[-1].offset += 100
        self.signal = s

    def test_align_zero_loss_peak_calibrate_true(self):
        s = self.signal
        s.align_zero_loss_peak(calibrate=True, print_stats=False)
        zlpc = s.estimate_zero_loss_peak_centre()
        np.testing.assert_allclose(zlpc.data.mean(), 0)
        np.testing.assert_allclose(zlpc.data.std(), 0)

    def test_align_zero_loss_peak_calibrate_true_with_mask(self):
        s = self.signal
        mask = s._get_navigation_signal(dtype="bool").T
        mask.data[[3, 5]] = (True, True)
        s.align_zero_loss_peak(calibrate=True, print_stats=False, mask=mask)
        zlpc = s.estimate_zero_loss_peak_centre(mask=mask)
        np.testing.assert_allclose(np.nanmean(zlpc.data), 0, atol=np.finfo(float).eps)
        np.testing.assert_allclose(np.nanstd(zlpc.data), 0, atol=np.finfo(float).eps)

    def test_align_zero_loss_peak_calibrate_false(self):
        s = self.signal
        s.align_zero_loss_peak(calibrate=False, print_stats=False)
        zlpc = s.estimate_zero_loss_peak_centre()
        np.testing.assert_allclose(zlpc.data.std(), 0, atol=10e-3)

    def test_also_aligns(self):
        s = self.signal
        s2 = s.deepcopy()
        s.align_zero_loss_peak(calibrate=True, print_stats=False, also_align=[s2])
        zlpc = s2.estimate_zero_loss_peak_centre()
        assert zlpc.data.mean() == 0
        assert zlpc.data.std() == 0

    def test_align_zero_loss_peak_with_spike_signal_range(self):
        s = self.signal
        spike = np.zeros((10, 100))
        spike_amplitude = 20
        spike[:, 75] = spike_amplitude
        s.data += spike
        s.align_zero_loss_peak(
            print_stats=False, subpixel=False, signal_range=(98.0, 102.0)
        )
        zlp_max = s.isig[-0.5:0.5].max(-1).data
        # Max value in the original spectrum is 12, but due to the aligning
        # the peak is split between two different channels. So 8 is the
        # maximum value for the aligned spectrum
        np.testing.assert_allclose(zlp_max, 8)

    def test_align_zero_loss_peak_crop_false(self):
        s = self.signal
        original_size = s.axes_manager.signal_axes[0].size
        s.align_zero_loss_peak(crop=False, print_stats=False)
        assert original_size == s.axes_manager.signal_axes[0].size

    @pytest.mark.parametrize("signal_range", ((-2.0, 2.0), (0, 40), "roi"))
    def test_align_zero_loss_peak_start_end_float(self, signal_range):
        s = self.signal
        if signal_range == "roi":
            signal_range = hs.roi.SpanROI(-3, 3)
        s.axes_manager[-1].offset = -2
        s.align_zero_loss_peak(subpixel=True, signal_range=signal_range)
        zlpc = s.estimate_zero_loss_peak_centre()
        # Check if start and end arguments work
        assert zlpc.data.mean() == 0
        assert zlpc.data.std() == 0


@lazifyTestClass
class TestSpikesRemovalToolZLP:
    def setup_method(self, method):
        # Create an empty spectrum
        s = exspy.signals.EELSSpectrum(np.zeros((2, 3, 64)))
        energy_axis = s.axes_manager.signal_axes[0]
        energy_axis.scale = 0.2
        energy_axis.offset = -5

        gauss = hs.model.components1D.Gaussian()
        gauss.centre.value = 0
        gauss.A.value = 5000
        gauss.sigma.value = 0.5
        s.data = s.data + gauss.function(energy_axis.axis)
        s.add_gaussian_noise(1, random_state=1)
        self.signal = s

    def _add_spikes(self):
        s = self.signal
        s.data[1, 0, 1] += 200
        s.data[0, 2, 29] += 500
        s.data[1, 2, 14] += 1000

    def test_get_zero_loss_peak_mask(self):
        mask = self.signal.get_zero_loss_peak_mask()
        expected_mask = np.zeros(self.signal.axes_manager.signal_size, dtype=bool)
        expected_mask[13:38] = True
        np.testing.assert_allclose(mask, expected_mask)

    def test_get_zero_loss_peak_mask_signal_mask(self):
        signal_mask = np.zeros(self.signal.axes_manager.signal_size, dtype=bool)
        signal_mask[40:50] = True
        mask = self.signal.get_zero_loss_peak_mask(signal_mask=signal_mask)
        expected_mask = np.zeros(self.signal.axes_manager.signal_size, dtype=bool)
        expected_mask[13:38] = True
        np.testing.assert_allclose(mask, np.logical_or(expected_mask, signal_mask))

    def test_spikes_diagnosis(self):
        if self.signal._lazy:
            pytest.skip("Spikes diagnosis is not supported for lazy signal")
        self._add_spikes()
        self.signal.spikes_diagnosis(zero_loss_peak_mask_width=5.0)

        zlp_mask = self.signal.get_zero_loss_peak_mask()
        hist_data = self.signal._spikes_diagnosis(signal_mask=zlp_mask, bins=25)
        expected_data = np.zeros(25)
        expected_data[0] = 232
        expected_data[12] = 1
        expected_data[-1] = 1
        np.testing.assert_allclose(hist_data.data, expected_data)

        hist_data2 = self.signal._spikes_diagnosis(bins=25)
        expected_data2 = np.array([286, 10, 13, 0, 0, 1, 12, 0])
        np.testing.assert_allclose(hist_data2.data[:8], expected_data2)

        # mask all to check that it raises an error when there is no data
        signal_mask = self.signal.inav[0, 1].data.astype(bool)
        with pytest.raises(ValueError):
            self.signal.spikes_diagnosis(signal_mask=signal_mask)


def test_spikes_removal_tool_no_zlp():
    s = exspy.data.EELS_MnFe()
    with pytest.raises(ValueError):
        # Zero not in energy range
        s.spikes_removal_tool(zero_loss_peak_mask_width=5.0)


def test_spikes_diagnosis_constant_derivative():
    s = hs.signals.Signal1D(np.arange(20).reshape(2, 10))
    with pytest.warns():
        s._spikes_diagnosis(use_gui=False)

    hs.preferences.GUIs.enable_traitsui_gui = False
    with pytest.warns():
        s._spikes_diagnosis(use_gui=True)

    hs.preferences.GUIs.enable_traitsui_gui = True
    try:
        import hyperspy_gui_traitsui

        s._spikes_diagnosis(use_gui=True)
    except ImportError:
        pass


@lazifyTestClass
class TestPowerLawExtrapolation:
    def setup_method(self, method):
        s = exspy.signals.EELSSpectrum(0.1 * np.arange(50, 250, 0.5) ** -3.0)
        s.axes_manager[-1].is_binned = False
        s.axes_manager[-1].offset = 50
        s.axes_manager[-1].scale = 0.5
        self.s = s

    def test_unbinned(self):
        sc = self.s.isig[:300]
        s = sc.power_law_extrapolation(extrapolation_size=100)
        np.testing.assert_allclose(s.data, self.s.data, atol=10e-3)

    def test_binned(self):
        self.s.data *= self.s.axes_manager[-1].scale
        self.s.axes_manager[-1].is_binned = True
        sc = self.s.isig[:300]
        s = sc.power_law_extrapolation(extrapolation_size=100)
        np.testing.assert_allclose(s.data, self.s.data, atol=10e-3)


@lazifyTestClass
class TestFourierRatioDeconvolution:
    @pytest.mark.parametrize(("extrapolate_lowloss"), [True, False])
    def test_running(self, extrapolate_lowloss):
        s = exspy.signals.EELSSpectrum(np.arange(200))
        gaussian = hs.model.components1D.Gaussian()
        gaussian.A.value = 50
        gaussian.sigma.value = 10
        gaussian.centre.value = 20
        s_ll = exspy.signals.EELSSpectrum(gaussian.function(np.arange(0, 200, 1)))
        s_ll.axes_manager[0].offset = -50
        s.fourier_ratio_deconvolution(s_ll, extrapolate_lowloss=extrapolate_lowloss)


@lazifyTestClass
class TestRebin:
    def setup_method(self, method):
        # Create an empty spectrum
        s = exspy.signals.EELSSpectrum(np.ones((4, 2, 1024)))
        self.signal = s

    def test_rebin_without_dwell_time(self):
        s = self.signal
        s.rebin(scale=(2, 2, 1))

    def test_rebin_dwell_time(self):
        s = self.signal
        s.metadata.add_node("Acquisition_instrument.TEM.Detector.EELS")
        s_mdEELS = s.metadata.Acquisition_instrument.TEM.Detector.EELS
        s_mdEELS.dwell_time = 0.1
        s_mdEELS.exposure = 0.5
        s2 = s.rebin(scale=(2, 2, 8))
        s2_mdEELS = s2.metadata.Acquisition_instrument.TEM.Detector.EELS
        assert s2_mdEELS.dwell_time == (0.1 * 2 * 2)
        assert s2_mdEELS.exposure == (0.5 * 2 * 2)

        def test_rebin_exposure(self):
            s = self.signal
            s.metadata.exposure = 10.2
            s2 = s.rebin(scale=(2, 2, 8))
            assert s2.metadata.exposure == (10.2 * 2 * 2)

    def test_offset_after_rebin(self):
        s = self.signal
        s.axes_manager[0].offset = 1
        s.axes_manager[1].offset = 2
        s.axes_manager[2].offset = 3
        s2 = s.rebin(scale=(2, 2, 1))
        assert s2.axes_manager[0].offset == 1.5
        assert s2.axes_manager[1].offset == 2.5
        assert s2.axes_manager[2].offset == s.axes_manager[2].offset


@lazifyTestClass
class Test_Estimate_Thickness:
    def setup_method(self, method):
        # Create an empty spectrum
        self.s = hs.load(
            MYPATH.joinpath("data/EELS_LL_linescan_simulated_thickness_variation.hspy")
        )
        self.zlp = hs.load(
            MYPATH.joinpath("data/EELS_ZLP_linescan_simulated_thickness_variation.hspy")
        )

    def test_relative_thickness(self):
        t = self.s.estimate_thickness(zlp=self.zlp)
        np.testing.assert_allclose(t.data, np.arange(0.3, 2, 0.1), atol=4e-3)
        assert t.metadata.Signal.quantity == "$\\frac{t}{\\lambda}$"

    def test_thickness_mfp(self):
        t = self.s.estimate_thickness(zlp=self.zlp, mean_free_path=120)
        np.testing.assert_allclose(t.data, 120 * np.arange(0.3, 2, 0.1), rtol=3e-3)
        assert t.metadata.Signal.quantity == "thickness (nm)"

    def test_thickness_density(self):
        t = self.s.estimate_thickness(zlp=self.zlp, density=3.6)
        np.testing.assert_allclose(t.data, 142 * np.arange(0.3, 2, 0.1), rtol=3e-3)
        assert t.metadata.Signal.quantity == "thickness (nm)"

    def test_thickness_density_and_mfp(self):
        t = self.s.estimate_thickness(zlp=self.zlp, density=3.6, mean_free_path=120)
        np.testing.assert_allclose(t.data, 127.5 * np.arange(0.3, 2, 0.1), rtol=3e-3)
        assert t.metadata.Signal.quantity == "thickness (nm)"

    def test_threshold(self):
        t = self.s.estimate_thickness(threshold=4.5, density=3.6, mean_free_path=120)
        np.testing.assert_allclose(t.data, 127.5 * np.arange(0.3, 2, 0.1), rtol=3e-3)
        assert t.metadata.Signal.quantity == "thickness (nm)"

    def test_threshold_nd(self):
        threshold = self.s._get_navigation_signal()
        threshold.data[:] = 4.5
        t = self.s.estimate_thickness(
            threshold=threshold, density=3.6, mean_free_path=120
        )
        np.testing.assert_allclose(t.data, 127.5 * np.arange(0.3, 2, 0.1), rtol=3e-3)
        assert t.metadata.Signal.quantity == "thickness (nm)"

    def test_no_zlp_or_threshold(self):
        with pytest.raises(ValueError):
            self.s.estimate_thickness()

    def test_no_metadata(self):
        del self.s.metadata.Acquisition_instrument
        with pytest.raises(RuntimeError):
            self.s.estimate_thickness(zlp=self.zlp, density=3.6)


class TestPrintEdgesNearEnergy:
    def setup_method(self, method):
        # Create an empty spectrum
        s = exspy.signals.EELSSpectrum(np.ones((4, 2, 1024)))
        self.signal = s

    def test_at_532eV(self, capsys):
        s = self.signal
        s.print_edges_near_energy(532)
        captured = capsys.readouterr()
        expected_out = (
            "+-------+-------------------+-----------+-----------------+\n"
            "|  edge | onset energy (eV) | relevance |   description   |\n"
            "+-------+-------------------+-----------+-----------------+\n"
            "|  O_K  |       532.0       |   Major   |   Abrupt onset  |\n"
            "| Pd_M3 |       531.0       |   Minor   |                 |\n"
            "| At_N5 |       533.0       |   Minor   |                 |\n"
            "| Sb_M5 |       528.0       |   Major   | Delayed maximum |\n"
            "| Sb_M4 |       537.0       |   Major   | Delayed maximum |\n"
            "+-------+-------------------+-----------+-----------------+\n"
        )
        assert captured.out == expected_out

    def test_sequence_edges(self, capsys):
        s = self.signal
        s.print_edges_near_energy(123, edges=["Mn_L2", "O_K", "Fe_L2"])
        captured = capsys.readouterr()
        expected_out = (
            "+-------+-------------------+-----------+-----------------------------+\n"
            "|  edge | onset energy (eV) | relevance |         description         |\n"
            "+-------+-------------------+-----------+-----------------------------+\n"
            "| Mn_L2 |       651.0       |   Major   | Sharp peak. Delayed maximum |\n"
            "|  O_K  |       532.0       |   Major   |         Abrupt onset        |\n"
            "| Fe_L2 |       721.0       |   Major   | Sharp peak. Delayed maximum |\n"
            "+-------+-------------------+-----------+-----------------------------+\n"
        )
        assert captured.out == expected_out

    def test_no_energy_and_edges(self):
        s = self.signal
        with pytest.raises(ValueError):
            s.print_edges_near_energy()


class Test_Edges_At_Energy:
    def setup_method(self, method):
        # Create an empty spectrum
        s = exspy.signals.EELSSpectrum(np.ones((4, 2, 1024)))
        self.signal = s

    def test_at_532eV(self, capsys):
        s = self.signal
        s.edges_at_energy(532, width=20, only_major=True, order="ascending")
        captured = capsys.readouterr()

        expected_out = (
            "+-------+-------------------+-----------+-----------------+\n"
            "|  edge | onset energy (eV) | relevance |   description   |\n"
            "+-------+-------------------+-----------+-----------------+\n"
            "| Sb_M5 |       528.0       |   Major   | Delayed maximum |\n"
            "|  O_K  |       532.0       |   Major   |   Abrupt onset  |\n"
            "| Sb_M4 |       537.0       |   Major   | Delayed maximum |\n"
            "+-------+-------------------+-----------+-----------------+\n"
        )
        assert captured.out == expected_out


class Test_Get_Complementary_Edges:
    def setup_method(self, method):
        # Create an empty spectrum
        s = exspy.signals.EELSSpectrum(np.ones((4, 2, 1024)))
        self.signal = s

    def test_Fe_O(self):
        s = self.signal
        complementary = s._get_complementary_edges(["Fe_L2", "O_K"])

        assert complementary == ["Fe_L1", "Fe_L3", "Fe_M3", "Fe_M2"]

    def test_Fe_O_only_major(self):
        s = self.signal
        complementary = s._get_complementary_edges(["Fe_L2", "O_K"], only_major=True)

        assert complementary == ["Fe_L3", "Fe_M3", "Fe_M2"]


class Test_Plot_EELS:
    def setup_method(self, method):
        s = exspy.data.EELS_MnFe()
        self.signal = s

    def test_plot_no_markers(self):
        s = self.signal
        s.add_elements(("Mn", "Cr"))
        s.plot()

        assert len(s._edge_markers["names"]) == 0

    def test_plot_edges_True(self):
        s = self.signal
        s.add_elements(("Mn", "Cr"))
        s.plot(plot_edges=True)

        print(s._edge_markers["names"])

        assert len(s._edge_markers["names"]) == 8
        assert set(s._edge_markers["names"]) == {
            "Cr_L2",
            "Cr_L3",
            "Cr_L1",
            "Fe_L2",
            "Fe_L3",
            "Mn_L2",
            "Mn_L3",
            "Mn_L1",
        }

    def test_plot_edges_True_without_elements(self):
        s = self.signal
        del s.metadata.Sample.elements
        s.metadata
        with pytest.raises(ValueError):
            s.plot(plot_edges=True)

    def test_plot_edges_from_element_family_specific(self):
        s = self.signal
        s.plot(plot_edges=["Mn", "Ti_L", "Cr_L3"], only_edges=("Major"))

        print(s._edge_markers["names"])

        assert len(s._edge_markers["names"]) == 7
        assert set(s._edge_markers["names"]) == {
            "Fe_L2",
            "Fe_L3",
            "Mn_L2",
            "Mn_L3",
            "Ti_L2",
            "Ti_L3",
            "Cr_L3",
        }

    def test_unsupported_edge_family(self):
        s = self.signal
        with pytest.raises(AttributeError):
            s.plot(plot_edges=["Cr_P"])

    def test_unsupported_edge(self):
        s = self.signal
        with pytest.raises(AttributeError):
            s.plot(plot_edges=["Xe_P4"])

    def test_unsupported_element(self):
        s = self.signal
        with pytest.raises(ValueError):
            s.plot(plot_edges=["ABC_L1"])

    def test_remove_edge_labels(self):
        s = self.signal
        del s.metadata.Sample.elements
        s.plot(plot_edges=["Cr_L", "Fe_L2"])
        s._remove_edge_labels(["Cr_L1", "Fe_L2"])

        assert len(s._edge_markers["names"]) == 2
        assert set(s._edge_markers["names"]) == set(["Cr_L2", "Cr_L3"])

    def test_plot_edges_without_markers_provided(self):
        s = self.signal
        s.plot()
        s._plot_edge_labels({"Fe_L2": 721.0, "O_K": 532.0})

        assert len(s._edge_markers["names"]) == 2
        assert set(s._edge_markers["names"]) == set(["Fe_L2", "O_K"])


@lazifyTestClass
class TestVacuumMask:
    def setup_method(self, method):
        s = exspy.signals.EELSSpectrum(np.array([np.linspace(0.001, 0.5, 20)] * 100).T)
        s.add_poissonian_noise(random_state=1)
        s.axes_manager[-1].scale = 0.25
        s.axes_manager[-1].units = "eV"
        s.inav[:10] += 20
        self.signal = s

    def test_vacuum_mask_opening(self):
        s = self.signal
        mask = s.vacuum_mask(opening=True)
        assert not mask.data[0]
        assert not mask.data[9]
        assert mask.data[10]
        assert mask.data[-1]

    def test_vacuum_mask_threshold(self):
        s = self.signal
        mask = s.vacuum_mask(threshold=20)
        assert mask.data[0]
        assert not mask.data[1]
        assert not mask.data[2]
        assert not mask.data[9]
        assert mask.data[10]
        assert mask.data[-1]

    def test_vacuum_mask_navigation_dimension_0(self):
        s = self.signal
        s2 = s.sum()
        with pytest.raises(RuntimeError):
            s2.vacuum_mask()



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import numpy as np
import pytest

from hyperspy.decorators import lazifyTestClass
from hyperspy.misc import utils

import exspy
from exspy.misc.eds import utils as utils_eds
from exspy.misc.elements import elements as elements_db


# Create this outside the test class to
# reduce computation in test suite by ~10seconds
s = utils_eds.xray_lines_model(
    elements=["Fe", "Cr", "Zn"],
    beam_energy=200,
    weight_percents=[20, 50, 30],
    energy_resolution_MnKa=130,
    energy_axis={
        "units": "keV",
        "size": 400,
        "scale": 0.01,
        "name": "E",
        "offset": 5.0,
    },
)
s = s + 0.002


@lazifyTestClass
class TestlineFit:
    def setup_method(self, method):
        self.s = s.deepcopy()

    def test_fit(self):
        s = self.s
        m = s.create_model()
        m.fit()
        np.testing.assert_allclose(
            [i.data for i in m.get_lines_intensity()], [[0.5], [0.2], [0.3]], atol=1e-4
        )

    def _check_model_creation(self):
        s = self.s
        # Default:
        m = s.create_model()
        assert [c.name for c in m] == [
            "background_order_6",
            "Cr_Ka",
            "Cr_Kb",
            "Fe_Ka",
            "Fe_Kb",
            "Zn_Ka",
        ]
        # No auto componentes:
        m = s.create_model(False, False)
        assert [c.name for c in m] == []

    def test_model_creation(self):
        self._check_model_creation()

    def test_semmodel_creation(self):
        self.s.set_signal_type("EDS_SEM")
        self._check_model_creation()

    def test_temmodel_creation(self):
        self.s.set_signal_type("EDS_TEM")
        self._check_model_creation()

    def _check_model_store(self):
        # Simply check that storing/restoring of EDSModels work
        # This also checks that creation from dictionary works
        s = self.s
        # Default:
        m = s.create_model()
        m.remove(["Cr_Ka", "background_order_6"])
        m.store()
        m1 = s.models.a.restore()
        assert [c.name for c in m] == [c.name for c in m1]
        assert [c.name for c in m.xray_lines] == [c.name for c in m1.xray_lines]
        assert "Cr_Ka" not in m1.xray_lines
        assert "background_order_6" not in m1.background_components

    def test_edsmodel_store(self):
        self._check_model_store()

    def test_semmodel_store(self):
        self.s.set_signal_type("EDS_SEM")
        self._check_model_store()

    def test_temmodel_store(self):
        self.s.set_signal_type("EDS_TEM")
        self._check_model_store()

    def test_calibrate_energy_resolution(self):
        s = self.s
        m = s.create_model()
        m.fit()
        m.fit_background()
        reso = (
            s.metadata.Acquisition_instrument.TEM.Detector.EDS.energy_resolution_MnKa,
        )
        s.set_microscope_parameters(energy_resolution_MnKa=150)
        m.calibrate_energy_axis(calibrate="resolution")
        np.testing.assert_allclose(
            s.metadata.Acquisition_instrument.TEM.Detector.EDS.energy_resolution_MnKa,
            reso,
            atol=1,
        )

    def test_calibrate_energy_scale(self):
        s = self.s
        m = s.create_model()
        m.fit()
        ax = s.axes_manager[-1]
        scale = ax.scale
        ax.scale += 0.0004
        m.calibrate_energy_axis("scale")
        np.testing.assert_allclose(ax.scale, scale, atol=1e-5)

    def test_calibrate_energy_offset(self):
        s = self.s
        m = s.create_model()
        m.fit()
        ax = s.axes_manager[-1]
        offset = ax.offset
        ax.offset += 0.04
        m.calibrate_energy_axis("offset")
        np.testing.assert_allclose(ax.offset, offset, atol=1e-5)

    def test_calibrate_xray_energy(self):
        s = self.s
        m = s.create_model()
        m.fit()
        m["Fe_Ka"].centre.value = 6.39

        m.calibrate_xray_lines(calibrate="energy", xray_lines=["Fe_Ka"], bound=100)

        np.testing.assert_allclose(
            m["Fe_Ka"].centre.value,
            elements_db["Fe"]["Atomic_properties"]["Xray_lines"]["Ka"]["energy (keV)"],
            atol=1e-6,
        )

    def test_calibrate_xray_weight(self):
        s = self.s
        s1 = utils_eds.xray_lines_model(
            elements=["Co"],
            weight_percents=[50],
            energy_axis={
                "units": "keV",
                "size": 400,
                "scale": 0.01,
                "name": "E",
                "offset": 4.9,
            },
        )
        s = s + s1 / 50
        m = s.create_model()
        m.fit()

        with pytest.warns(
            UserWarning, match="X-ray line expected to be in the model was not found"
        ):
            m.calibrate_xray_lines(
                calibrate="sub_weight", xray_lines=["Fe_Ka"], bound=100
            )

        np.testing.assert_allclose(0.0347, m["Fe_Kb"].A.value, atol=1e-3)

    def test_calibrate_xray_width(self):
        s = self.s
        m = s.create_model()
        m.fit()
        sigma = m["Fe_Ka"].sigma.value
        m["Fe_Ka"].sigma.value = 0.065

        m.calibrate_xray_lines(calibrate="energy", xray_lines=["Fe_Ka"], bound=10)

        np.testing.assert_allclose(sigma, m["Fe_Ka"].sigma.value, atol=1e-2)

    def test_enable_adjust_position(self):
        m = self.s.create_model()
        m.enable_adjust_position()
        assert len(m._position_widgets) == 5
        # Check that both line and label was added
        assert len(list(m._position_widgets.values())[0]) == 2
        lbls = [p[1].string for p in m._position_widgets.values()]
        assert sorted(lbls) == [
            "$\\mathrm{Cr}_{\\mathrm{Ka}}$",
            "$\\mathrm{Cr}_{\\mathrm{Kb}}$",
            "$\\mathrm{Fe}_{\\mathrm{Ka}}$",
            "$\\mathrm{Fe}_{\\mathrm{Kb}}$",
            "$\\mathrm{Zn}_{\\mathrm{Ka}}$",
        ]

    def test_quantification(self):
        s = self.s
        # to have enough intensities, so that the default values for
        # `navigation_mask` of 1.0 in `quantification` doesn't mask the data
        s = s * 1000
        m = s.create_model()
        m.fit()
        intensities = m.get_lines_intensity()
        quant = s.quantification(
            intensities,
            method="CL",
            factors=[1.0, 1.0, 1.0],
            composition_units="weight",
        )
        np.testing.assert_allclose(utils.stack(quant, axis=0), [50, 20, 30])

    def test_quantification_2_elements(self):
        s = self.s
        m = s.create_model()
        m.fit()
        intensities = m.get_lines_intensity(["Fe_Ka", "Cr_Ka"])
        _ = s.quantification(intensities, method="CL", factors=[1.0, 1.0])


def test_comparison_quantification():
    kfactors = [1.450226, 5.075602]  # For Fe Ka and Pt La

    s = exspy.data.EDS_TEM_FePt_nanoparticles()
    s.add_elements(["Cu"])  # to get good estimation of the background
    m = s.create_model(True)
    m.set_signal_range(5.5, 10.0)  # to get good fit
    m.fit()
    intensities_m = m.get_lines_intensity(xray_lines=["Fe_Ka", "Pt_La"])
    quant_model = s.quantification(intensities_m, method="CL", factors=kfactors)

    # Background substracted EDS quantification
    s2 = exspy.data.EDS_TEM_FePt_nanoparticles()
    s2.add_lines()
    bw = s2.estimate_background_windows(
        xray_lines=["Fe_Ka", "Pt_La"], line_width=[5.0, 2.0]
    )
    intensities = s2.get_lines_intensity(
        xray_lines=["Fe_Ka", "Pt_La"], background_windows=bw
    )
    atomic_percent = s2.quantification(intensities, method="CL", factors=kfactors)

    np.testing.assert_allclose(
        [q.data for q in quant_model], [q.data for q in atomic_percent], rtol=5e-2
    )


@lazifyTestClass
class TestMaps:
    def setup_method(self, method):
        beam_energy = 200
        energy_resolution_MnKa = 130
        energy_axis = {"units": "keV", "size": 1200, "scale": 0.01, "name": "E"}
        s1 = utils_eds.xray_lines_model(
            elements=["Fe", "Cr"],
            weight_percents=[30, 70],
            beam_energy=beam_energy,
            energy_resolution_MnKa=energy_resolution_MnKa,
            energy_axis=energy_axis,
        )
        s2 = utils_eds.xray_lines_model(
            elements=["Ga", "As"],
            weight_percents=[50, 50],
            beam_energy=beam_energy,
            energy_resolution_MnKa=energy_resolution_MnKa,
            energy_axis=energy_axis,
        )

        mix = np.linspace(0.0, 1.0, 4).reshape(2, 2)
        mix_data = np.tile(s1.data, mix.shape + (1,))
        s = s1._deepcopy_with_new_data(mix_data)
        a = s.axes_manager._axes.pop(0).get_axis_dictionary()
        s.axes_manager.create_axes([{"size": mix.shape[0], "navigate": True}] * 2 + [a])
        s.add_elements(s2.metadata.Sample.elements)

        for d, m in zip(s._iterate_signal(iterpath="flyback"), mix.flatten()):
            d[:] = d * m + (1 - m) * s2.data
        self.mix = mix
        self.s = s

    def test_lines_intensity(self):
        s = self.s

        m = s.create_model()
        m.multifit()
        ws = np.array([0.5, 0.7, 0.3, 0.5])
        w = np.zeros((4,) + self.mix.shape)
        for x in range(self.mix.shape[0]):
            for y in range(self.mix.shape[1]):
                for i in range(4):
                    mix = self.mix[x, y] if i in (1, 2) else 1 - self.mix[x, y]
                    w[i, x, y] = ws[i] * mix
        xray_lines = s._get_lines_from_elements(
            s.metadata.Sample.elements, only_lines=("Ka",)
        )
        if s._lazy:
            s.compute()
        for fitted, expected in zip(m.get_lines_intensity(xray_lines), w):
            np.testing.assert_allclose(fitted, expected, atol=1e-7)

        m_single_fit = s.create_model()
        # make sure we fit the navigation indices (0, 0) and don't rely on
        # the current index of the axes_manager.
        m_single_fit.inav[0, 0].fit()

        for fitted, expected in zip(
            m.inav[0, 0].get_lines_intensity(xray_lines),
            m_single_fit.inav[0, 0].get_lines_intensity(xray_lines),
        ):
            np.testing.assert_allclose(fitted, expected, atol=1e-7)



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The HyperSpy developers
#
# This file is part of HyperSpy.
#
# HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import numpy as np

from exspy.components import VolumePlasmonDrude


def test_function():
    g = VolumePlasmonDrude()
    g.intensity.value = 3.0
    g.plasmon_energy.value = 8.0
    g.fwhm.value = 2.0
    assert g.function(0) == 0
    np.testing.assert_allclose(g.function(8), 12)
    np.testing.assert_allclose(g.function(1), 9.66524e-2, rtol=1e-6)
    np.testing.assert_allclose(g.function(30), 1.639867e-2, rtol=1e-6)
